/* Actions for killing users.
 * src/actions.c
 * Copyright (C) 1999-2000 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef ACTIONS_C
#define ACTIONS_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

/* kill_user()
 * Remove a user from the IRC network.  source is the nick which should
 * generate the kill, or NULL for a server generated kill. */

void kill_user(const char *source, const char *user, const char *reason)
{
   char *av[2];

   if (!user || !*user)
      return;
   if (!source || !*source)
      source = sstrdup(conf.lserver);
   if (!reason)
      reason = "";
   av[0] = sstrdup(user);
   av[1] = smalloc(strlen(source) + strlen(reason) + 4);
   sprintf(av[1], "%s (%s)", source, reason);
   send_cmd(source, "KILL %s :%s", user, av[1]);
   do_kill(source, 2, av);
   free(av[0]);
   free(av[1]);
}

/*************************************************************************/

/* bad_password()
 * Note a bad password attempt for the given user.  If they've used up
 * their limit, remove them from the network. */

void bad_password(User *u)
{
   time_t now;

   if (!conf.nsbadpassword)
      return;
   else
      now = time(NULL);
   if (conf.nsbadtimeout > 0) {
      if (u->invalid_pw_time > 0 && u->invalid_pw_time < now - conf.nsbadtimeout)
         u->invalid_pw_count = 0;
   }
   u->invalid_pw_count++;
   u->invalid_pw_time = now;
   if (u->invalid_pw_count >= conf.nsbadpassword)
      kill_user(NULL, u->nick, "Too many invalid passwords");
}

/*************************************************************************/

#endif

/*************************************************************************/