/* Channel functionss.
 * src/channels.c
 * Copyright (C) 1999-2000 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef CHANNELS_C
#define CHANNELS_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

static Channel *chanlist = NULL;

/*************************************************************************/

/* get_channel_stats()
 * Return the memory usage for active channels. */

void get_channel_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    Channel *chan;
    struct c_userlist *cu;
    int i, j;

      for (chan = chanlist; chan; chan = chan->next) {
	    count++;
	    mem += sizeof(*chan);
	    if (chan->topic)
		mem += strlen(chan->topic)+1;
	    if (chan->key)
		mem += strlen(chan->key)+1;
	    mem += sizeof(char *) * chan->bansize;
	    for (j = 0; j < chan->bancount; j++) {
		if (chan->bans[j])
		    mem += strlen(chan->bans[j])+1;
	    }
	    for (cu = chan->users; cu; cu = cu->next)
		mem += sizeof(*cu);
	    for (cu = chan->chanops; cu; cu = cu->next)
		mem += sizeof(*cu);
	    for (cu = chan->voices; cu; cu = cu->next)
		mem += sizeof(*cu);
#ifdef V_UNREAL
	    for (cu = chan->halfops; cu; cu = cu->next)
		mem += sizeof(*cu);
#endif
	}
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* Return the Channel structure corresponding to the named channel, or NULL
 * if the channel was not found.  chan is assumed to be non-NULL and valid
 * (i.e. pointing to a channel name of 2 or more characters). */

Channel *findchan (const char *chan)
{
    Channel *c = chanlist;

    for (c = chanlist; c; c = c->next)
	if (stricmp (c->name, chan) == 0)
	    return c;
    return NULL;
}

/*************************************************************************/

Channel *chan_create(const char *chan) 
{
   Channel *c = findchan(chan);
   int newchan = !c;
   struct c_userlist *u;

   /* Allocate pre-cleared memory */
   if (newchan) {
	c = scalloc(sizeof(Channel), 1);
	c->next = chanlist;
	if (chanlist)
		chanlist->prev = c;
	chanlist = c;
	strscpy(c->name, chan, sizeof(c->name));
	c->creation_time = time(NULL);
	c->ci = cs_findchan(chan);
	if (c->ci)
	    c->ci->c = c;
	check_modes(chan);
	restore_topic(chan);
   }
   return c;

}

/* chan_adduser()
 * Add/remove a user to/from a channel, creating or deleting the channel as
 * necessary. */

void chan_adduser(User *user, const char *chan)
{
    Channel *c = findchan(chan);

    if (!c) c = chan_create(chan);
    chan_adduser2(user, c);
}

/* OLD METHOD */

/*
void chan_adduser(User *user, const char *chan)
{
    Channel *c = findchan(chan);
    int newchan = !c;
    struct c_userlist *u;

    if (newchan) {
	c = scalloc(sizeof(Channel), 1);
	c->next = chanlist;
	if (chanlist)
		chanlist->prev = c;
	chanlist = c;
	strscpy(c->name, chan, sizeof(c->name));
	c->creation_time = time(NULL);
	c->ci = cs_findchan(chan);
	if (c->ci)
	    c->ci->c = c;
	check_modes(chan);
	restore_topic(chan);
    }
    if (check_should_op(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->chanops;
	u->prev = NULL;
	if (c->chanops)
	    c->chanops->prev = u;
	c->chanops = u;
	u->user = user;
#ifdef V_UNREAL
    } else if (check_should_halfop(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->halfops;
	u->prev = NULL;
	if (c->halfops)
	    c->halfops->prev = u;
	c->halfops = u;
	u->user = user;
#endif
    } else if (check_should_voice(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->voices;
	u->prev = NULL;
	if (c->voices)
	    c->voices->prev = u;
	c->voices = u;
	u->user = user;
    } else {
    }

    u = smalloc(sizeof(struct c_userlist));
    u->next = c->users;
    u->prev = NULL;
    if (c->users)
	c->users->prev = u;
    c->users = u;
    u->user = user;
}
*/

void chan_adduser2(User *user, Channel *c)
{
    struct c_userlist *u;
    char *chan = c->name;	/* FIXME: remove this in the near future ;) */
    
    if (!c) return;

    if (check_should_op(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->chanops;
	u->prev = NULL;
	if (c->chanops)
	    c->chanops->prev = u;
	c->chanops = u;
	u->user = user;
#ifdef V_UNREAL
    } else if (check_should_halfop(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->halfops;
	u->prev = NULL;
	if (c->halfops)
	    c->halfops->prev = u;
	c->halfops = u;
	u->user = user;
#endif
    } else if (check_should_voice(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->voices;
	u->prev = NULL;
	if (c->voices)
	    c->voices->prev = u;
	c->voices = u;
	u->user = user;
    } else if (check_should_srop(user, chan)) {
	u = smalloc(sizeof(struct c_userlist));
	u->next = c->chanops;
	u->prev = NULL;
	if (c->chanops)
	    c->chanops->prev = u;
	c->chanops = u;
	u->user = user;
    } else {
    }

    u = smalloc(sizeof(struct c_userlist));
    u->next = c->users;
    u->prev = NULL;
    if (c->users)
	c->users->prev = u;
    c->users = u;
    u->user = user;

}

/*************************************************************************/

/* chan_deluser()
 * Remove a user from a channel, if last user remove the channel. */

void chan_deluser(User *user, Channel *c)
{
    struct c_userlist *u;

    for (u = c->users; u && u->user != user; u = u->next)
	;
    if (!u)
	return;
    if (u->next)
	u->next->prev = u->prev;
    if (u->prev)
	u->prev->next = u->next;
    else
	c->users = u->next;
    free(u);
    for (u = c->chanops; u && u->user != user; u = u->next)
	;
    if (u) {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->chanops = u->next;
	free(u);
    }
#ifdef V_UNREAL
    for (u = c->halfops; u && u->user != user; u = u->next)
	;
    if (u) {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->halfops = u->next;
	free(u);
    }
#endif
    for (u = c->voices; u && u->user != user; u = u->next)
	;
    if (u) {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->voices = u->next;
	free(u);
    }
    if (!c->users) {
	if (c->ci)
	    c->ci->c = NULL;
	if (c->topic)
	    free(c->topic);
	if (c->key)
	    free(c->key);
	if (c->bancount) {
	    int i;
	    for (i = 0; i < c->bancount; ++i)
		free(c->bans[i]);
	}
	if (c->bansize)
	    free(c->bans);

#ifdef V_UNREAL
	if (c->chanops || c->voices || c->halfops)
         log("channel: Memory leak freeing %s, possibly channel members (()->users, ()->chanops, ()->halfops, ()->voices).", c->name);
#else
	if (c->chanops || c->voices)
         log("channel: Memory leak freeing %s, possibly channel members (()->users, ()->chanops, ()->voices).", c->name);
#endif
	if (c->next)
	    c->next->prev = c->prev;
	if (c->prev)
	    c->prev->next = c->next;
	else
		chanlist = c->next;
	free(c);
    }
}

/*************************************************************************/

/* free_all_chan()
 * Frees all the channel structure entires. */

void free_all_chan(void)
{
   Channel *c;
   struct c_userlist *u;
   int i;

   for (c = chanlist; c; c = c->next) {
	   for (u = c->users; u ; u = u->next)
		free(u);
	   for (u = c->voices; u ; u = u->next)
		free(u);
	   for (u = c->chanops; u ; u = u->next)
		free(u);
#ifdef V_UNREAL
	   for (u = c->halfops; u ; u = u->next)
		free(u);
#endif
	   if (!c->users) {
		if (c->ci)
		    c->ci->c = NULL;
		if (c->topic)
		    free(c->topic);
		if (c->key)
		    free(c->key);
		if (c->bancount) {
		    for (i = 0; i < c->bancount; ++i)
			free(c->bans[i]);
		}
		if (c->bansize)
		    free(c->bans);
		/* if (c->next)
		    c->next->prev = c->prev;
		if (c->prev)
		    c->prev->next = c->next;
		else
			chanlist = c->next; */
		free(c);
    	   }
   }
}

/*************************************************************************/

/* Handle a channel MODE command. */

void do_cmode(const char *source, int ac, char **av)
{
    Channel *chan;
    ChannelInfo *ci;
    struct c_userlist *u;
    User *user;
    char *s, *nick;
    int add = 1;		/* 1 if adding modes, 0 if deleting */
    char *modestr = av[1];
    int source_is = 0, nick_is = 0;

    chan = findchan(av[0]);
    if (!chan) {
	log("channel: MODE %s for nonexistent channel %s",
					merge_args(ac-1, av+1), av[0]);
	return;
    }
    s = modestr;
    ac -= 2;
    av += 2;

    while (*s) {

	switch (*s++) {

	case '+':
	    add = 1; break;

	case '-':
	    add = 0; break;

#ifdef V_BAHAMUT
	case 'c':
	    if (add)
		chan->mode |= CMODE_C;
	    else
		chan->mode &= ~CMODE_C;
	    break;
#endif
	case 'i':
	    if (add)
		chan->mode |= CMODE_I;
	    else
		chan->mode &= ~CMODE_I;
	    break;

	case 'm':
	    if (add)
		chan->mode |= CMODE_M;
	    else
		chan->mode &= ~CMODE_M;
	    break;

	case 'n':
	    if (add)
		chan->mode |= CMODE_N;
	    else
		chan->mode &= ~CMODE_N;
	    break;

	case 'p':
	    if (add)
		chan->mode |= CMODE_P;
	    else
		chan->mode &= ~CMODE_P;
	    break;

	case 's':
	    if (add)
		chan->mode |= CMODE_S;
	    else
		chan->mode &= ~CMODE_S;
	    break;

	case 't':
	    if (add)
		chan->mode |= CMODE_T;
	    else
		chan->mode &= ~CMODE_T;
	    break;

	case 'k':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %ck",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (chan->key) {
		free(chan->key);
		chan->key = NULL;
	    }
	    if (add)
		chan->key = sstrdup(*av++);
	    break;

	case 'l':
	    if (add) {
		if (--ac < 0) {
		    log("channel: MODE %s %s: missing parameter for +l",
							chan->name, modestr);
		    break;
		}
		chan->limit = atoi(*av++);
	    } else {
		chan->limit = 0;
	    }
	    break;

	case 'b':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %cb",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    if (add) {
		if (chan->bancount >= chan->bansize) {
		    chan->bansize += 8;
		    chan->bans = srealloc(chan->bans, sizeof(char *) * chan->bansize);
		}
		chan->bans[chan->bancount++] = sstrdup(*av++);
	    } else {
		char **s = chan->bans;
		int i = 0;
		while (i < chan->bancount && strcmp(*s, *av) != 0) {
		    i++;
		    s++;
		}
		if (i < chan->bancount) {
		    chan->bancount--;
		    if (i < chan->bancount)
			bcopy(s+1, s, sizeof(char *) * (chan->bancount-i));
		}
		av++;
	    }
	    break;

	case 'o':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %co",
					chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    nick = *av++;
	    if (add) {
		for (u = chan->chanops; u && stricmp(u->user->nick, nick) != 0;
								u = u->next)
		    ;
		if (u)
		    break;
		user = finduser(nick);
		if (!user) {
		    log("channel: MODE %s +o for nonexistent user %s",
							chan->name, nick);
		    break;
		}
		if (!check_valid_op(user, chan->name, !!strchr(source, '.')))
		    break;
		u = smalloc(sizeof(*u));
		u->next = chan->chanops;
		u->prev = NULL;
		if (chan->chanops)
		    chan->chanops->prev = u;
		chan->chanops = u;
		u->user = user;
	    } else {
		/* If OpGuard is on, we need to do t he following:
             * a) is 'nick' an AOp Or SOp
             * b) if so is 'source' AOp or SOp
             * c) If not, reop 'nick' and deop source 
             * Services clients, services roots and related clients are exempt from this check. */
		if ((ci = cs_findchan(chan->name)) && (ci->flags & CI_OPGUARD)) {
               if (!is_services_client(source) && !is_client(source) && !is_services_root(source)) {
			if (is_chan_aop(nick, ci) || is_chan_sop(nick, ci))
				nick_is = 1;
			if (is_chan_aop(source, ci) || is_chan_sop(source, ci))
				source_is = 1;
			if (nick_is && !source_is) {
				send_cmd(s_ChanServ, "MODE %s +o-o %s %s", chan->name, nick, source);
				for (u = chan->chanops; u && stricmp(u->user->nick, source); u = u->next)
				    ;
				if (!u)
				    break;
				if (u->next)
				    u->next->prev = u->prev;
				if (u->prev)
				    u->prev->next = u->next;
				else
				    chan->chanops = u->next;
				free(u);
			}
		   }
		} else {
			for (u = chan->chanops; u && stricmp(u->user->nick, nick); u = u->next)
			    ;
			if (!u)
			    break;
			if (u->next)
			    u->next->prev = u->prev;
			if (u->prev)
			    u->prev->next = u->next;
			else
			    chan->chanops = u->next;
			free(u);
		}
  	   }
	    break;
	case 'v':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %cv", chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    nick = *av++;
	    if (add) {
		for (u = chan->voices; u && stricmp(u->user->nick, nick); u = u->next)
		    ;
		if (u)
		    break;
		user = finduser(nick);
		if (!user) {
		    log("channe: MODE %s +v for nonexistent user %s", chan->name, nick);
		    break;
		}
		u = smalloc(sizeof(*u));
		u->next = chan->voices;
		u->prev = NULL;
		if (chan->voices)
		    chan->voices->prev = u;
		chan->voices = u;
		u->user = user;
	    } else {
		for (u = chan->voices; u && stricmp(u->user->nick, nick); u = u->next)
		    ;
		if (!u)
		    break;
		if (u->next)
		    u->next->prev = u->prev;
		if (u->prev)
		    u->prev->next = u->next;
		else
		    chan->voices = u->next;
		free(u);
	    }
	    break;
#ifdef V_UNREAL
	case 'h':
	    if (--ac < 0) {
		log("channel: MODE %s %s: missing parameter for %ch", chan->name, modestr, add ? '+' : '-');
		break;
	    }
	    nick = *av++;
	    if (add) {
		for (u = chan->halfops; u && stricmp(u->user->nick, nick); u = u->next)
		    ;
		if (u)
		    break;
		user = finduser(nick);
		if (!user) {
		    log("channe: MODE %s +h for nonexistent user %s", chan->name, nick);
		    break;
		}
		u = smalloc(sizeof(*u));
		u->next = chan->halfops;
		u->prev = NULL;
		if (chan->halfops)
		    chan->halfops->prev = u;
		chan->halfops = u;
		u->user = user;
	    } else {
		for (u = chan->halfops; u && stricmp(u->user->nick, nick); u = u->next)
		    ;
		if (!u)
		    break;
		if (u->next)
		    u->next->prev = u->prev;
		if (u->prev)
		    u->prev->next = u->next;
		else
		    chan->halfops = u->next;
		free(u);
	    }
	    break;
#endif
	} /* switch */

    } /* while (*s) */

    /* Check modes against ChanServ mode lock */
    check_modes(chan->name);
}

/*************************************************************************/

/* Handle a TOPIC command. */

void do_topic(const char *source, int ac, char **av)
{
    Channel *c = findchan(av[0]);

    if (!c) {
	log("channel: TOPIC %s for nonexistent channel %s", merge_args(ac-1, av+1), av[0]);
	return;
    }
    if (check_topiclock(av[0]))
	return;
    strscpy(c->topic_setter, av[1], sizeof(c->topic_setter));
    c->topic_time = atol(av[2]);
    if (c->topic) {
	free(c->topic);
	c->topic = NULL;
    }
    if (ac > 3 && *av[3])
	c->topic = sstrdup(av[3]);

   if (conf.chanscan) {
    if (match_wild_nocase("*IRC.*", av[3]) || match_wild_nocase("*/SERVER*", av[3]) || 
	  match_wild_nocase("*MOVING*", av[3])) {
	if (conf.spcchannelactive)
	      privmsg(s_ChanServ, conf.spcchannel, "Channel %s has a topic set matching a possible spam, server or moving channel criteria -> (%s).", 
			av[0], av[3]);
    } else if (match_wild_nocase("*ATTACK*", av[3]) || match_wild_nocase("*DOS*", av[3]) || 
	  match_wild_nocase("*PACKET*", av[3])) {
	if (conf.spcchannelactive)
	      privmsg(s_ChanServ, conf.spcchannel, "Channel %s has a topic set matching a possible war, dos or packet channel criteria -> (%s).", 
			av[0], av[3]);
    } else if (match_wild_nocase("*DAL.NET*", av[3]) || match_wild_nocase("*RELIC.NET*", av[3]) || 
	  match_wild_nocase("*STARCHAT.NET*", av[3]) || match_wild_nocase("*INSIDERZ.NET*", av[3]) || 
	  match_wild_nocase("*DARKFIRE.NET*", av[3]) || match_wild_nocase("*NEWNET.NET*", av[3]) ||
	  match_wild_nocase("*UNDERNET.ORG*", av[3]) || match_wild_nocase("*IRCNET*", av[3]) ||
	  match_wild_nocase("*EFNET*", av[3])) {
	if (conf.spcchannelactive)
	      privmsg(s_ChanServ, conf.spcchannel, "Channel %s has a topic set matching a possible network or server advertisment channel criteria -> (%s).", 
			av[0], av[3]);
    }
   }
    record_topic(av[0]);
}

/*************************************************************************/

/* only_one_user()
 * Does the given channel have only one user? */

int only_one_user(const char *chan)
{
    Channel *c = findchan(chan);
    return (c && c->users && !c->users->next) ? 1 : 0;
}

/*************************************************************************/

#endif

/*************************************************************************/
