/* Configuration file functions
 * src/conf.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef CONF_C
#define CONF_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

static int parse_conf 				FUNCTION((char *option, char *value));
static void conf_set_nicks 			FUNCTION((void));

/*************************************************************************/

/* Are we rehashing the configuration file? */
int rehash_conf;

/* Please note, the numerical assigments in this structure are crucial to
 * the loading of the configuration file, you must increment 2,3,4,5 not
 * 10,20,24, same applies for parsing the options and values. */

typedef struct {
   int pos;
   char option[CONF_OPTION_MAX];
   char value[CONF_VALUE_MAX];
} Options;
Options options[] = {
   { 0,          "Undefined",          "Undefined"         },
   { 1,          "Include",            "Undefined"         },
   { 2,          "RemoteServer",       "Undefined"         },
   { 3,          "RemotePort",         "Undefined"         },
   { 4,          "LinkPassword",       "Undefined"         },
   { 5,          "LocalServer",        "Undefined"         },
   { 6,          "LocalDesc",          "Undefined"         },
   { 7,          "LogFile",            "services.log"      },
   { 8,          "PidFile",            "services.pid"      },
   { 9,          "MotdFile",           "services.motd"     },
   { 10,         "CsNick",             "ChanServ"          },
   { 11,         "CsIdent",            "chanserv"          },
   { 12,         "CsHost",             "services.lyth.org" },
   { 13,         "CsModes",            "+oi-fcgwsaA"       },
   { 14,         "CsDb",               "chan.db"           },
   { 15,         "CsAKicks",           "64"                },
   { 16,         "Undefined",          ""                  },
   { 17,         "CsExpiry",           "14"                },
   { 18,         "CsAKickReason",      "Banned"            },
   { 19,         "CsInhabit",          "15"                },
   { 20,         "NsNick",             "NickServ"          },
   { 21,         "NsIdent",            "nickserv"          },
   { 22,         "NsHost",             "services.lyth.org" },
   { 23,         "NsModes",            "+oi-fcgwsaA"       },
   { 24,         "NsDb",               "nick.db"           },
   { 25,         "NsRegDelay",         "30"                },
   { 26,         "NsExpiry",           "30"                },
   { 27,         "NsRelease",          "60"                },
   { 28,         "MsNick",             "MemoServ"          },
   { 29,         "MsIdent",            "memoserv"          },
   { 30,         "MsHost",             "services.lyth.org" },
   { 31,         "MsModes",            "+oi-fcgwsaA"       },
   { 32,         "MsDb",               "memo.db"           },
   { 33,         "Undefined",          ""                  },
   { 34,         "MsSndDelay",         "15"                },
   { 35,         "OsNick",             "OperServ"          },
   { 36,         "OsIdent",            "operserv"          },
   { 37,         "OsHost",             "services.lyth.org" },
   { 38,         "OsModes",            "+oi-fcgwsaA"       },
   { 39,         "OsAdminDb",          "admin.db"          },
   { 40,         "OsAkillDb",          "akill.db"          },
   { 41,         "OsRoot",             "Root"              },
   { 42,         "OsCheckClones",      "1"                 },
   { 43,         "OsClones",           "100"               },
   { 44,         "OsMinClones",        "5"                 },
   { 45,         "OsDelay",            "10"                },
   { 46,         "OsWarnDelay",        "30"                },
   { 47,         "OsAkillMsg",         "akill@lyth.org"    },
   { 48,         "HsNick",             "HelpServ"          },
   { 49,         "HsIdent",            "helpserv"          },
   { 50,         "HsHost",             "services.lyth.org" },
   { 51,         "HsModes",            "+io-fcgwsaA"       },
   { 52,         "GnNick",             "Lyth"              },
   { 53,         "GnIdent",            "lyth"              },
   { 54,         "GnHost",             "services.lyth.org" },
   { 55,         "GnModes",            "+io-fcgwsaA"       },
   { 56,         "NsStrict",           "1"                 },
   { 57,         "NsListOper",         "1"                 },
   { 58,         "NsBadPassword",      "5"                 },
   { 59,         "NsBadTimeout",       "3600"              },
   { 60,         "DumpChannel",        "#dsnoop"	     },
   { 61,         "DumpChannelActive",  "1"                 },
   { 62,         "DumpChannelModes",   "+tnisr"            },
   { 63,         "ServicesActive",     "1"                 },
   { 64,         "CsActive",           "1"                 },
   { 65,         "NsActive",           "1"                 },
   { 66,         "MsActive",           "1"                 },
   { 67,         "OsActive",           "1"                 },
   { 68,         "HsActive",           "1"                 },
   { 69,         "LogonMsg1",          ""                  },
   { 70,         "LogonMsg2",          ""                  },
   { 71,         "LogonMsg3",          ""                  },
   { 72,         "OperMsg1",           ""                  },
   { 73,         "SendPings",          "1"                 },
   { 74,         "SendPingsTimeout",   "90"                },
   { 75,         "OperMsg2",           ""                  },
   { 76,         "OperMsg3",           ""                  },
   { 77,         "MsMailBinary",       ""                  },
   { 78,         "MsMailFooter",       ""                  },
   { 79,         "MsMailSubject"       ""                  },
   { 80,         "MsMailActive",       ""                  }, 
   { 81,         "NsRegisterMemo",     ""                  },
   { 82,         "KlineEmail",         ""                  },
   { 83,         "KlineUrl",           ""                  },
   { 84,         "FloodMsgs",          ""                  },
   { 85,         "FloodSMsgs",         ""                  },
   { 86,         "FloodMsgsTime",      ""                  },
   { 87,         "OsRouteDb",          ""                  },
   { 88,         "GetRoot",            ""                  },
   { 89,         "MsCDb",              ""                  },
   { 90,         "ServScan",           "1"                 },
   { 91,         "ChanScan",           "1"                 },
   { 92,         "SpcChannel",         "#dsspc"	     },
   { 93,         "SpcChannelActive",   "1"                 },
   { 94,         "SpcChannelModes",    "+tnisr"            },
   { 95,         "YLineUNum",          "1"     		     },
   { 96,         "YLineUMax",  	       "5000"              },
   { 97,         "YLineUPing",         "90"                },
   { 98,         "YLineUSendQ",        "2000"     	     },
   { 99,         "YLineONum",          "2"     		     },
   { 100,        "YLineOMax",  	       "100"               },
   { 101,        "YLineOPing",         "90"                },
   { 102,        "YLineOSendQ",        "5000"     	     },
   { 103,        "MsExpiry",           "5"                },

   /* Expanision, increment by 1, all the time. */
   { -1 }
};
    
/*************************************************************************/

/* init_conf()
 * Opens the configuration file then calls parse_conf() to parse the options and the values
 * from the configuration file. */

int load_conf(char *conf) 
{
   FILE *f;
   char buffer[BUFSIZE];
   char option[CONF_OPTION_MAX];
   char value[CONF_VALUE_MAX]; 
   char *cp1, *cp2;
   int parse_conf_status;
   int i, optcount = 0;

   /* Open the configuration file for reading. */
   if ((f = fopen(conf, "r")) == NULL) {
      return 0;
   }

   while ((fgets(buffer, BUFSIZE, f)) != NULL) {
      /* Comments are # and ;, plus [] support for sections. */
      if ((buffer[0] == '#') || (buffer[0] == ';') || (buffer[0] == '\n') || (buffer[0] == '[' && buffer[strlen(buffer)] == ']'))
         continue;

      /* Get an option from input buffer. */
      cp1 = buffer;
      cp2 = option;
      while (isalnum(*cp1))
         *cp2++ = *cp1++;
      *cp2 = '\0';

      /* Get a value from the input buffer. */
      cp2 = value;
      while ((*cp1 != '\n') && (*cp1 != '\0') && (isspace(*cp1)))
         cp1++;
      while ((*cp1 != '\n') && (*cp1 != '\0'))
         *cp2++ = *cp1++;
      *cp2-- = '\0';
      while ((isspace(*cp2)) && (cp2 != value)) 
         *cp2-- = '\0';

      if (!(parse_conf(option, value)))
         continue;
      /* Since we had a sucessful option, update the conf options counter */
      optcount++;
   }
   fclose(f);

   return 1;
}

/*************************************************************************/

/* parse_conf()
 * Parses the configuration file for its options and values and stores them in the conf
 * structure. */

static int parse_conf(char *option, char *value)
{
   int i, j, key = 0;
   char *rhost;

   /* Check if blank or null option or value. */
   if ((option[0] == '\0') || (value[0] == '\0'))
      return 0;

   for (i = 0; options[i].pos >= 0; i++)
      if (stricmp(option, options[i].option) == 0)
         key = i;

   /* Option is not reconginzed, so ignore it. */
   if (key == 0)
      return 0;

   /* Proccess all options and their values. */
   switch (key) {
      case 1:
         if (load_conf(value))
            break;
      case 2: strscpy(conf.rserver, value, HOSTMAX);                      break;
      case 3: conf.rport = atoi(value);                                   break;
      case 4: strscpy(conf.lpassword, value, PASSMAX);                    break;
      case 5: strscpy(conf.lserver, value, HOSTMAX);                      break;
      case 6: strscpy(conf.ldesc, value, DESCMAX);	                  break;
      case 7: conf.logfile = sstrdup(value);                              break;
      case 8: conf.pidfile = sstrdup(value);                              break;
      case 9: conf.motdfile = sstrdup(value);                             break;
      case 10: strscpy(conf.csnick, value, NICKMAX);                      break;
      case 11: strscpy(conf.csident, value, IDENTMAX);                    break;
      case 12: strscpy(conf.cshost, value, HOSTMAX);                      break;
      case 13: strscpy(conf.csmodes, value, MODEMAX);                     break;
      case 14: strscpy(conf.chandb, value, DBMAX);		          break;
      case 15: conf.csakicks = atoi(value);                               break;
      case 17: conf.csexpiry = atoi(value);                               break;
      case 18: strscpy(conf.csakickreason, value, MSGMAX);	          break;
      case 19: conf.csinhabit = atoi(value);                              break;
      case 20: strscpy(conf.nsnick, value, NICKMAX);                      break;
      case 21: strscpy(conf.nsident, value, IDENTMAX);                    break;
      case 22: strscpy(conf.nshost, value, HOSTMAX);                      break;
      case 23: strscpy(conf.nsmodes, value, MODEMAX);                     break;
      case 24: strscpy(conf.nickdb, value, DBMAX);		          break;
      case 25: conf.nsregdelay = atoi(value);                             break;
      case 26: conf.nsexpiry = atoi(value);                               break;
      case 27: conf.nsrelease = atoi(value);                              break;
      case 28: strscpy(conf.msnick, value, NICKMAX);                      break;
      case 29: strscpy(conf.msident, value, IDENTMAX);                    break;
      case 30: strscpy(conf.mshost, value, HOSTMAX);                      break;
      case 31: strscpy(conf.msmodes, value, MODEMAX);                     break;
      case 32: strscpy(conf.memodb, value, DBMAX);		          break;
      case 34: conf.mssnddelay = atoi(value);                             break;
      case 35: strscpy(conf.osnick, value, NICKMAX);                      break;
      case 36: strscpy(conf.osident, value, IDENTMAX);                    break;
      case 37: strscpy(conf.oshost, value, HOSTMAX);                      break;
      case 38: strscpy(conf.osmodes, value, MODEMAX);                     break;
      case 39: strscpy(conf.admindb, value, DBMAX);		          break;
      case 40: strscpy(conf.akilldb, value, DBMAX);		          break;
      case 41: strscpy(conf.osroot, value, NICKMAX);                      break;
      case 47: strscpy(conf.osakillmsg, value, MSGMAX);		          break;
      case 48: strscpy(conf.hsnick, value, NICKMAX);                      break;
      case 49: strscpy(conf.hsident, value, IDENTMAX);                    break;
      case 50: strscpy(conf.hshost, value, HOSTMAX);                      break;
      case 51: strscpy(conf.hsmodes, value, MODEMAX);                     break;
      case 52: strscpy(conf.gnnick, value, NICKMAX);                      break;
      case 53: strscpy(conf.gnident, value, IDENTMAX);                    break;
      case 54: strscpy(conf.gnhost, value, HOSTMAX);                      break;
      case 55: strscpy(conf.gnmodes, value, MODEMAX);                     break;
      case 56: conf.nsstrict = atoi(value);                               break;
      case 57: conf.nslistoper = atoi(value);                             break;
      case 58: conf.nsbadpassword = atoi(value);                          break;
      case 59: conf.nsbadtimeout = atoi(value);                           break;
      case 60: strscpy(conf.dumpchannel, value, CHANMAX);                 break;
      case 61: conf.dumpchannelactive = atoi(value);                      break;
      case 62: strscpy(conf.dumpchannelmodes, value, MODEMAX);            break;
      case 63: conf.servicesactive = atoi(value);                         break;
      case 64: conf.csactive = atoi(value);                               break;
      case 65: conf.nsactive = atoi(value);                               break;
      case 66: conf.msactive = atoi(value);                               break;
      case 67: conf.osactive = atoi(value);                               break;
      case 68: conf.hsactive = atoi(value);                               break;
      case 69: strscpy(conf.logonmsg1, value, MSGMAX);                    break;
      case 70: strscpy(conf.logonmsg2, value, MSGMAX);                    break;
      case 71: strscpy(conf.logonmsg3, value, MSGMAX);                    break;
      case 72: strscpy(conf.opermsg1, value, MSGMAX);                     break;
      case 73: conf.sendpings = atoi(value);                              break; 
      case 74: conf.sendpingstimeout = atoi(value);                       break; 
      case 75: strscpy(conf.opermsg2, value, MSGMAX);                     break;
      case 76: strscpy(conf.opermsg3, value, MSGMAX);                     break;
      case 77: conf.msmailbinary = sstrdup(value);                        break;
      case 78: conf.msmailfooter = sstrdup(value);                        break;
      case 79: conf.msmailsubject = sstrdup(value);                       break;
      case 80: conf.msmailactive = atoi(value);                           break;
      case 81: strscpy(conf.nsregistermemo, value, MSGMAX);               break;
      case 82: strscpy(conf.klineemail, value, HOSTMAX);          	  break;
      case 83: strscpy(conf.klineurl, value, HOSTMAX);                    break;
      case 84: conf.floodmsgs = atoi(value);                              break; 
      case 85: conf.floodsmsgs = atoi(value);                             break; 
      case 86: conf.floodmsgstime = atoi(value);                          break; 
      case 87: strscpy(conf.routedb, value, DBMAX);		          break;
      case 88: strscpy(conf.getroot, value, PASSMAX);		          break;
      case 89: strscpy(conf.cmemodb, value, DBMAX);		          break;
      case 90: conf.servscan = atoi(value);                               break; 
      case 91: conf.chanscan = atoi(value);                               break; 
      case 92: strscpy(conf.spcchannel, value, CHANMAX);                  break;
      case 93: conf.spcchannelactive = atoi(value);                       break;
      case 94: strscpy(conf.spcchannelmodes, value, MODEMAX);             break;
      case 95: conf.yunum = atoi(value);                                  break; 
      case 96: conf.yumax = atoi(value);                                  break; 
      case 97: conf.yuping = atoi(value);                                 break; 
      case 98: conf.yusendq = atoi(value);                                break; 
      case 99: conf.yonum = atoi(value);                                  break; 
      case 100: conf.yomax = atoi(value);                                 break; 
      case 101: conf.yoping = atoi(value);                                break; 
      case 102: conf.yosendq = atoi(value);                               break; 
      case 103: conf.msexpiry = atoi(value);                              break;
      default: break;
   }
   /* Options and their values were added or ignored since they already exist. */
   return 1;
}

/*************************************************************************/

/* setup_var_from_conf()
 * Setup the variables need for services with this function, also used in rehashing. */

void setup_var_from_conf(void)
{
   if (rehash_conf) {
      /* Send quit notice to clients. */
      send_cmd(s_NickServ, "QUIT");
      send_cmd(s_ChanServ, "QUIT");
      send_cmd(s_MemoServ, "QUIT");
      send_cmd(s_OperServ, "QUIT");
      send_cmd(s_HelpServ, "QUIT");
      send_cmd(s_GlobalNoticer, "QUIT");
      conf_set_nicks();
      /* Introduce the new clients. */
      introduce_user(NULL);
	   if (conf.dumpchannelactive)
		join_channel(conf.dumpchannel);
	   if (conf.spcchannelactive)
		join_channel(conf.spcchannel);
      rehash_conf = 0;
   } else {
      conf_set_nicks();
   }
}

/*************************************************************************/

/* conf_set_nicks()
 * Set the new nicks for services after rehashing. */

static void conf_set_nicks(void)
{
   if (conf.nsnick)
      strscpy(s_NickServ, conf.nsnick, NICKMAX);
   if (conf.csnick)
      strscpy(s_ChanServ, conf.csnick, NICKMAX);
   if (conf.msnick)
      strscpy(s_MemoServ, conf.msnick, NICKMAX);
   if (conf.osnick)
      strscpy(s_OperServ, conf.osnick, NICKMAX);
   if (conf.hsnick)
      strscpy(s_HelpServ, conf.hsnick, NICKMAX);
   if (conf.gnnick)
      strscpy(s_GlobalNoticer, conf.gnnick, NICKMAX);
}

/*************************************************************************/

#endif

/*************************************************************************/