/* MemoServ functions.
 * src/memoserv.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to this software. */

/*************************************************************************/

#ifndef MEMOSERV_C
#define MEMOSERV_C

/*************************************************************************/

#include "services.h"
#include "ms-help.c"

/*************************************************************************/

static MemoList *find_memolist 			FUNCTION((const char *nick));
static CMemoList *find_cmemolist 			FUNCTION((const char *chan));
static void alpha_insert_memolist 			FUNCTION((MemoList *ml));
static void del_memolist 				FUNCTION((MemoList *ml));
static void alpha_insert_cmemolist 			FUNCTION((CMemoList *cml));
static void del_cmemolist 				FUNCTION((CMemoList *cml));
static void ms_help  					FUNCTION((const char *source));
static void ms_csend 					FUNCTION((const char *source));
static void ms_clist 					FUNCTION((const char *source));
static void ms_cread 					FUNCTION((const char *source));
static void ms_cdel  					FUNCTION((const char *source));
static void ms_send  					FUNCTION((const char *source));
static void ms_msend 					FUNCTION((const char *source));
static void ms_list  					FUNCTION((const char *source));
static void ms_read  					FUNCTION((const char *source));
static void ms_del   					FUNCTION((const char *source));
static void ms_undel   					FUNCTION((const char *source));
static void ms_reply 					FUNCTION((const char *source));
static void ms_forward 					FUNCTION((const char *source));
static void ms_rlist 					FUNCTION((const char *source));
static void ms_rread 					FUNCTION((const char *source));
static void ms_rdel 					FUNCTION((const char *source));
static void ms_mmemo 					FUNCTION((const char *source));
static void ms_mail 					FUNCTION((const char *source));
static void ms_set 					FUNCTION((const char *source));
static void ms_opersend 				FUNCTION((const char *source));
static void ms_usersend 				FUNCTION((const char *source));
static void ms_adminsend 				FUNCTION((const char *source));
static void ms_set_notify 				FUNCTION((NickInfo *ni, char *param));

/*************************************************************************/

static MemoList *memolists[256];
static CMemoList *cmemolists[256];

/*************************************************************************/

/* get_memoserv_stats()
 * Return information on memory use.  Assumes pointers are valid. */

void get_memoserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    MemoList *ml;
    CMemoList *cml;

    crash_id = 65;

    for (i = 0; i < 256; i++) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    count++;
	    mem += sizeof(*ml);
	   if (ml->n_memos) {
		    mem += sizeof(Memo) * ml->n_memos;
		    for (j = 0; j < ml->n_memos; j++)
			if (ml->memos[j].text)
			    mem += strlen(ml->memos[j].text)+1;
	   }
	   if (ml->n_mail) {
		    mem += sizeof(Memo) * ml->n_mail;
		    for (j = 0; j < ml->n_mail; j++)
			if (ml->mail[j].text)
			    mem += strlen(ml->mail[j].text)+1;
	   }
	   if (ml->n_trash) {
		   mem += sizeof(Memo) * ml->n_trash;
		   for (j = 0; j < ml->n_trash; j++)
			if (ml->trash[j].text)
			   mem += strlen(ml->trash[j].text)+1;
	   }
	}
   }
    for (i = 0; i < 256; i++) {
	for (cml = cmemolists[i]; cml; cml = cml->next) {
	    count++;
	    mem += sizeof(*cml);
	   if (cml->n_trash) {
		    mem += sizeof(Memo) * cml->n_trash;
		    for (j = 0; j < cml->n_trash; j++)
			if (cml->trash[j].text)
			    mem += strlen(cml->trash[j].text)+1;
	   }
	   if (cml->n_memos) {
		    mem += sizeof(Memo) * cml->n_memos;
		    for (j = 0; j < cml->n_trash; j++)
			if (cml->trash[j].text)
			    mem += strlen(cml->trash[j].text)+1;
	   }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* memoserv()
 * Main MemoServ command processing routine. */

void memoserv(const char *source, char *buf)
{
   char *cmd, *s;
   User *u = finduser(source);
   NickInfo *ni;   
   char slvl[32];
   cmd = strtok(buf, " ");

   crash_id = 66;

   if (!cmd) {
      return;
   } else if (stricmp(cmd, "\1PING") == 0) {
      if (!(s = strtok(NULL, "")))
	   s = "\1";
      notice(s_MemoServ, source, "\1PING %s", s);
   } else {
	Hash *ms, ms_hash_table[] = {
	    { "HELP",		SA_NONE, 	ms_help,		H_NIDENTIFY	},
	    { "READ",		SA_NONE, 	ms_read,		H_IDENTIFY	},
	    { "SEND", 		SA_NONE, 	ms_send,		H_IDENTIFY	},
	    { "MSEND", 		SA_NONE, 	ms_msend,		H_IDENTIFY	},
	    { "LIST", 		SA_NONE, 	ms_list,		H_IDENTIFY	},
	    { "DEL", 		SA_NONE, 	ms_del,		H_IDENTIFY	},
	    { "REPLY", 		SA_NONE, 	ms_reply,		H_IDENTIFY	},
	    { "FORWARD", 		SA_NONE, 	ms_forward,		H_IDENTIFY	},
	    { "CSEND", 		SA_NONE, 	ms_csend,		H_IDENTIFY	},
	    { "CLIST", 		SA_NONE, 	ms_clist,		H_IDENTIFY	},
	    { "CDEL", 		SA_NONE, 	ms_cdel,		H_IDENTIFY	},
	    { "CREAD", 		SA_NONE, 	ms_cread,		H_IDENTIFY	},
	    { "SET", 		SA_NONE, 	ms_set,		H_IDENTIFY	},
	    { "MAIL",		SA_NONE, 	ms_mail,		H_IDENTIFY	},
	    { "OPERSEND",		SA_OPER, 	ms_opersend,	H_IDENTIFY	},
	    { "ADMINSEND", 	SA_ADMIN, 	ms_adminsend,	H_IDENTIFY	},
	    { "USERSEND",		SA_ADMIN, 	ms_usersend,	H_IDENTIFY	},
	    { "MMEMO", 		SA_SOP, 	ms_mmemo,		H_IDENTIFY	},
	    { "RLIST",		SA_SR, 	ms_rlist,		H_IDENTIFY	},
	    { "RREAD",		SA_SR, 	ms_rread,		H_IDENTIFY	},
	    { "RDEL",		SA_SR, 	ms_rdel,		H_IDENTIFY	},
	    { NULL }
	};
	if (ms = do_hash(source, cmd, ms_hash_table)) {
	   switch (get_admin_level(source)) {
	       case 0: strscpy(slvl, "IRC Operator", 32); 			break;
	       case 1: strscpy(slvl, "Services Administrator", 32); 	break;
	       case 2: strscpy(slvl, "Services Operator", 32); 		break;
	       case 3: strscpy(slvl, "Services Root", 32); 			break;
		 default: break;
	   }
         if ((ni = findnick(source)) && (ni->flags & NI_FROZEN)) {
             notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_IS_FROZEN));
	   } else if (ni && (ms->identify == H_IDENTIFY) && !(ni->flags & NI_IDENTIFIED)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PASSWORD_AUTH_REQUIRED));
             notice(s_MemoServ, source, rpl_numeric_lang(source, CR_RETRY_AFTER_TYPE), s_NickServ);
	   } else if (!ni && (ms->identify == H_IDENTIFY)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_YOUR_NICK_NOT_REGISTERED));
	   } else if (ms->level == SA_OPER && !is_oper(source)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NOT_OPER_LEVEL), "IRC Operator", "IRC Operator");
	   } else if (ms->level == SA_ADMIN && !is_services_admin(source)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Administrator.");
	   } else if (ms->level == SA_SOP && !is_services_op(source)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Operator.");
	   } else if (ms->level >= SA_SR && !is_services_root(source)) {
	       notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Root.");
	   } else {
		(*ms->func) (source);
	   }		
	} else {
	   notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_UNKNOWN), cmd, s_MemoServ);
      }
   }
}

/*************************************************************************/

/* load_ms_dbase()
 * Load the MemoServ data from it's database. */

void load_ms_dbase(void)
{
    FILE *f = fopen(conf.memodb, "r"), *ff = fopen("cmemo.db", "r");
    int ver, i, j;
    MemoList *ml;
    CMemoList *cml;
    Memo *memos, *mail, *trash, *cmemos, *ctrash;

    crash_id = 67;

    if (!f || !ff)
       return;
    for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {

		ml = smalloc(sizeof(MemoList));
		if (1 != fread(ml, sizeof(MemoList), 1, f))
		    fatal_perror("Read error on %s", conf.memodb);
		alpha_insert_memolist(ml);
		if (ml->n_memos) {
			ml->memos = memos = smalloc(sizeof(Memo) * ml->n_memos);
			fread(memos, sizeof(Memo), ml->n_memos, f);
			for (j = 0; j < ml->n_memos; j++, memos++)
			    memos->text = read_string(f, conf.memodb);
		}
		if (ml->n_mail) {
			ml->mail = mail = smalloc(sizeof(Memo) * ml->n_mail);
			fread(mail, sizeof(Memo), ml->n_mail, f);
			for (j = 0; j < ml->n_mail; j++, mail++)
			    mail->text = read_string(f, conf.memodb);
		}
		if (ml->n_trash) {
			ml->trash = trash = smalloc(sizeof(Memo) * ml->n_trash);
			fread(trash, sizeof(Memo), ml->n_trash, f);
			for (j = 0; j < ml->n_trash; j++, trash++)
			    trash->text = read_string(f, conf.memodb);
		}
	  }
    }
    fclose(f);
 	for (i = 33; i < 256; ++i) {
	    while (fgetc(ff) == 1) {

		cml = smalloc(sizeof(CMemoList));
		if (1 != fread(cml, sizeof(CMemoList), 1, ff))
		    fatal_perror("Read error on %s", "cmemo.db");
		alpha_insert_cmemolist(cml);
		if (cml->n_memos) {
			cml->memos = cmemos = smalloc(sizeof(Memo) * cml->n_memos);
			fread(cmemos, sizeof(Memo), cml->n_memos, ff);
			for (j = 0; j < cml->n_memos; j++, cmemos++)
			    cmemos->text = read_string(ff, "cmemo.db");
		}
		if (cml->n_trash) {
			cml->trash = ctrash = smalloc(sizeof(Memo) * cml->n_trash);
			fread(ctrash, sizeof(Memo), cml->n_trash, ff);
			for (j = 0; j < cml->n_trash; j++, ctrash++)
			    ctrash->text = read_string(ff, "cmemo.db");
		}
	    }
	}
   fclose(ff);
}

/*************************************************************************/

/* save_ms_dbase()
 * Save the MemoServ data to it's database. */

void save_ms_dbase(void)
{
    FILE *f = fopen(conf.memodb, "w"), *ff = fopen("cmemo.db", "w");
    int i, j;
    MemoList *ml;
    CMemoList *cml;
    Memo *memos, *mail, *trash, *cmemos, *ctrash;
    time_t save_start, save_end;

    if (!f || !ff)
       return;
   save_start = time(NULL);
   globops(conf.lserver, "Performing memo database save...");
    for (i = 33; i < 256; i++) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    fputc(1, f);
	    if (1 != fwrite(ml, sizeof(MemoList), 1, f)) 
		fatal_perror("Write error on %s", conf.memodb);

          if (ml->n_memos != fwrite(ml->memos, sizeof(Memo), ml->n_memos, f))
		fatal_perror("Write error on %s", conf.memodb);
		if (ml->n_memos) {
		    for (memos = ml->memos, j = 0; j < ml->n_memos; memos++, j++)
				write_string(memos->text, f, conf.memodb);
		}
	    if (ml->n_mail != fwrite(ml->mail, sizeof(Memo), ml->n_mail, f))
		fatal_perror("Write error on %s", conf.memodb);
		if (ml->n_mail) {
		    for (mail = ml->mail, j = 0; j < ml->n_mail; mail++, j++)
				write_string(mail->text, f, conf.memodb);
		}
          if (ml->n_trash != fwrite(ml->trash, sizeof(Memo), ml->n_trash, f))
		fatal_perror("Write error on %s", conf.memodb);
		if (ml->n_trash) {
		    for (trash = ml->trash, j = 0; j < ml->n_trash; trash++, j++)
				write_string(trash->text, f, conf.memodb);
		}
	}
	fputc(0, f);
    }
    fclose(f);
    for (i = 33; i < 256; i++) {
	for (cml = cmemolists[i]; cml; cml = cml->next) {
	    fputc(1, ff);
	    if (1 != fwrite(cml, sizeof(CMemoList), 1, ff)) 
		fatal_perror("Write error on %s", "cmemo.db");
          if (cml->n_memos != fwrite(cml->memos, sizeof(Memo), cml->n_memos, ff))
		fatal_perror("Write error on %s", "cmemo.db");

		if (cml->n_memos) {
		    for (cmemos = cml->memos, j = 0; j < cml->n_memos; cmemos++, j++)
				write_string(cmemos->text, ff, "cmemo.db");
		}
          if (cml->n_trash != fwrite(cml->trash, sizeof(Memo), cml->n_trash, ff))
		fatal_perror("Write error on %s", "cmemo.db");

		if (cml->n_trash) {
		    for (ctrash = cml->trash, j = 0; j < cml->n_trash; ctrash++, j++)
				write_string(ctrash->text, ff, "cmemo.db");
		}
       }
	fputc(0, ff);
    }
   fclose(ff);
   save_end = time(NULL) - save_start;
   globops(conf.lserver, "Memo database save completed in %d seconds.", save_end);
}

/*************************************************************************/

/* check_memos()
 * See if the given nick has any waiting memos, and send a NOTICE to that
 * nick if so. */

void check_memos(const char *nick)
{
    NickInfo *ni;
    MemoList *ml;
    int i, newcnt = 0, pri = 0, inbox_cnt = 0, mail_cnt = 0;

    crash_id = 68;

    if (!(ni = findnick(nick)) || !(ni->flags & NI_MEMO_SIGNON))
	return;
    if (ml = find_memolist(nick)) {

	for (i = 0; i < ml->n_memos; i++) {
	    if (ml->memos[i].flags & MF_UNREAD && ml->memos[i].flags & MF_L_INBOX) {
		newcnt++;
		inbox_cnt++;
	    }
	}
	for (i = 0; i < ml->n_mail; i++) {
	    if (ml->mail[i].flags & MF_UNREAD && ml->mail[i].flags & MF_L_MAIL) {
		newcnt++;
		mail_cnt++;
	    }
	}
	if (newcnt > 0) {
          notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_CHECK_MEMOS_NEW_MEMOS),
			newcnt, newcnt == 1 ? "" : "s", inbox_cnt, mail_cnt);
          notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_CHECK_MEMOS_PLEASE_TYPE),
			s_MemoServ, newcnt == 1 ? "READ [INBOX|MAIL] LAST" : "LIST [INBOX|MAIL] NEW",
			newcnt == 1 ? "read it" : "list them");
	}
    }
}

/*************************************************************************/

/* expire_memos() */

void expire_memos()
{
    const time_t expire_time = (5 * 24 * 60 * 60);
    int i, j, k ,l, total_memos = 0, expired_memos = 0, total_cmemos = 0, expired_cmemos = 0;
    time_t expire_start, expire_end, now = time(NULL);
    NickInfo *ni, *nnext;
    ChannelInfo *ci, *cnext;
    MemoList *ml;
    CMemoList *cml;
    Memo *m;

    crash_id = 40;

    expire_start = time(NULL);
    globops(conf.lserver, "Performing memo database expire...");
    if (5 > 0) {
       for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = nnext) {
	       nnext = ni->next;
		 if ((ml = find_memolist(ni->nick))) {
		    for (j = 0, m = ml->trash; j < ml->n_trash; ++j, ++m) {
		       if (m->flags & MF_L_TRASH) {
			    total_memos++;
			    if (now - m->time >= expire_time) {
			       if (j < ml->n_trash) {
				    free(ml->trash[j].text); 		
				    ml->n_trash--;		 		
			  	    if (j < ml->n_trash)
				       bcopy(ml->trash + j+1, ml->trash + j, sizeof(Memo) * (ml->n_trash - j));
			          expired_memos++;
				 }
			    }
			 }
		    }
		 }
          }
       }
       for (i = 0; i < 256; i++) {
	    for (ci = chanlists[k]; ci; ci = cnext) {
	       cnext = ci->next;
		 if ((cml = find_cmemolist(ci->name))) {
		    for (l = 0, m = cml->trash; l < cml->n_trash; ++l, ++m) {
		       if (m->flags & MF_L_TRASH) {
			    total_cmemos++;
			    if (now - m->time >= expire_time) {
			       if (l < cml->n_trash) {
				    free(cml->trash[l].text); 		
				    cml->n_trash--;		 		
			  	    if (l < cml->n_trash)
				       bcopy(cml->trash + l+1, cml->trash + l, sizeof(Memo) * (cml->n_trash - l));
			          expired_cmemos++;
				 }
			    }
			 }
		    }
		 }
          }
       }
    }
   expire_end = time(NULL) - expire_start;
   globops(conf.lserver, "Memo database expire completed NE(%d/%d) CE(%d/%d) in %d seconds.", 
     expired_memos, total_memos, expired_cmemos, total_cmemos, expire_end);
}

/*************************************************************************/

/* check_cmemos()
 * See if the given nick has any waiting memos, and send a NOTICE to that
 * nick if so. */

void check_cmemos(User *u, const char *chan)
{
    ChannelInfo *ci;
    CMemoList *cml;
    int i, newcnt = 0, inbox_cnt = 0, memo_cnt = 0;

    crash_id = 68;

    if (!(ci = cs_findchan(chan)))
	return;
    /* Only notice those who have the ability to read channel memos that there are memos for the target
     * channel. */
    if (check_access(u, ci, CA_READ_MEMO)) {
       if (cml = find_cmemolist(chan)) {
	   for (i = 0; i < cml->n_memos; i++) {
	       if (cml->memos[i].flags & MF_UNREAD && cml->memos[i].flags & MF_L_INBOX) {
		   newcnt++;
		   inbox_cnt++;
	       }
	       memo_cnt++;
	   }
         /* Always send a notice of how many memos or new memos there is, for users joining channels. */
	   if (newcnt > 0) {
             notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_CHECK_CMEMOS_NEW_MEMOS),
		   ci->name, newcnt, newcnt == 1 ? "" : "s", inbox_cnt);
             notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_CHECK_CMEMOS_PLEASE_TYPE),
		   s_MemoServ, newcnt == 1 ? "CREAD <channel> INBOX LAST" : "LIST <channel> INBOX NEW",
		   newcnt == 1 ? "read it" : "list them");
	   } else {
             notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_CHECK_CMEMOS_MEMOS),
		   ci->name, memo_cnt, memo_cnt == 1 ? "" : "s", memo_cnt);
             notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_CHECK_CMEMOS_PLEASE_TYPE),
		   s_MemoServ, memo_cnt == 1 ? "CREAD <channel> INBOX LAST" : "LIST <channel> INBOX",
		   memo_cnt == 1 ? "read it" : "list them");
	   }
      }
   }
}

/*************************************************************************/

/* clear_memos()
 * Clear all memos (if any) for a given user. */

void clear_memos(const char *nick)
{
    MemoList *ml;

    crash_id = 69;

    if (ml = find_memolist(nick))
	del_memolist(ml);
}

/*************************************************************************/

/* clear_cmemos()
 * Clear all memos (if any) for a given channel. */

void clear_cmemos(const char *chan)
{
    CMemoList *cml;

    crash_id = 69;

    if (cml = find_cmemolist(chan))
	del_cmemolist(cml);
}

/*************************************************************************/

/* find_memolist()
 * Find the memo list for a given nick.  Return NULL if none. */

static MemoList *find_memolist(const char *nick)
{
    MemoList *ml;
    int i = 1;

    crash_id = 70;

    for (ml = memolists[tolower(*nick)]; ml && (i = stricmp(ml->nick, nick)) < 0; ml = ml->next)
	;
    return i==0 ? ml : NULL;
}

/*************************************************************************/

/* alpha_insert_memolist()
 * Insert a memo list alphabetically into the database. */

static void alpha_insert_memolist(MemoList *ml)
{
    MemoList *ptr, *prev;
    char *nick = ml->nick;

    crash_id = 71;

    for (prev = NULL, ptr = memolists[tolower(*nick)]; ptr && stricmp(ptr->nick, nick) < 0; prev = ptr, ptr = ptr->next)
	;
    ml->prev = prev;
    ml->next = ptr;
    if (!prev)
	memolists[tolower(*nick)] = ml;
    else
	prev->next = ml;
    if (ptr)
	ptr->prev = ml;
}

/*************************************************************************/

/* del_memolist()
 * Remove a nick's memo list from the database. */

static void del_memolist(MemoList *ml)
{
    crash_id = 72;

    if (ml->next)
	ml->next->prev = ml->prev;
    if (ml->prev)
	ml->prev->next = ml->next;
    else
	memolists[tolower(*ml->nick)] = ml->next;
    if (ml->memos)
	free(ml->memos);
    if (ml->mail)
	free(ml->mail);
    if (ml->trash)
	free(ml->trash);
    free(ml);
}

/*************************************************************************/

/* find_cmemolist()
 * Find the memo list for a given channel.  Return NULL if none. */

static CMemoList *find_cmemolist(const char *chan)
{
    CMemoList *cml;
    int i = 1;

    crash_id = 70;

    for (cml = cmemolists[tolower(chan[1])]; cml && (i = stricmp(cml->chan, chan)) < 0; cml = cml->next)
	;
    return i == 0 ? cml : NULL;
}

/*************************************************************************/

/* alpha_insert_cmemolist()
 * Insert a memo list alphabetically into the database. */

static void alpha_insert_cmemolist(CMemoList *cml)
{
    CMemoList *ptr, *prev;
    char *chan = cml->chan;

    crash_id = 71;

    for (prev = NULL, ptr = cmemolists[tolower(chan[1])]; ptr && stricmp(ptr->chan, chan) < 0; prev = ptr, ptr = ptr->next)
	;
    cml->prev = prev;
    cml->next = ptr;
    if (!prev)
	cmemolists[tolower(chan[1])] = cml;
    else
	prev->next = cml;
    if (ptr)
	ptr->prev = cml;
}

/*************************************************************************/

/* del_cmemolist()
 * Remove a channel's memo list from the database. */

static void del_cmemolist(CMemoList *cml)
{
    crash_id = 72;

    if (cml->next)
	cml->next->prev = cml->prev;
    if (cml->prev)
	cml->prev->next = cml->next;
    else
	cmemolists[tolower(cml->chan[1])] = cml->next;
    if (cml->memos)
	free(cml->memos);
    if (cml->trash)
	free(cml->trash);
    free(cml);
}

/*************************************************************************/

/* ms_help()
 * Return a help message. */

static void ms_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];

    crash_id = 73;

    /* Show IRC Operator and higher commands to their own level. */
    if (cmd && is_oper(source)) {
       if (stricmp(cmd, "OPERSEND") == 0) {
	    notice_list(s_MemoServ, source, mso_opersend_help);
	    return;
       } else if (stricmp(cmd, "ADMINSEND") == 0 && is_services_admin(source)) {
	    notice_list(s_MemoServ, source, mso_adminsend_help);
	    return;
       } else if (stricmp(cmd, "USERSEND") == 0 && is_services_op(source)) {
	    notice_list(s_MemoServ, source, mso_usersend_help);
	    return;
       } else if (stricmp(cmd, "RREAD") == 0 && is_services_root(source)) {
	    notice_list(s_MemoServ, source, mso_rread_help);
	    return;
       } else if (stricmp(cmd, "RDEL") == 0 && is_services_root(source)) {
	    notice_list(s_MemoServ, source, mso_rdel_help);
	    return;
       } else if (stricmp(cmd, "RLIST") == 0 && is_services_root(source)) {
	    notice_list(s_MemoServ, source, mso_rlist_help);
	    return;
       }
    }
    /* Try to find the specified help file, if none exist, no help error pops up. */
    snprintf(buf, sizeof(buf), "%s%s", s_MemoServ, cmd ? " " : "");
    strscpy(buf + strlen(buf), cmd ? cmd : "", sizeof(buf) - strlen(buf));
    helpserv(s_MemoServ, source, buf);
    /* Show end of help for index to users who wont make it to the next !cmd check. */
    if (!cmd && !is_oper(source))
	notice(s_MemoServ, source, "\2***** End Of Help *****\2");

    /* No command, this will list all commands for the current level. */
    if (!cmd && is_oper(source)) {
	notice_list(s_MemoServ, source, ms_serv_oper_help);
	if (is_services_admin(source))
		notice_list(s_MemoServ, source, ms_serv_admin_help);
	if (is_services_op(source))
		notice_list(s_MemoServ, source, ms_serv_op_help);
	if (is_services_root(source))
		notice_list(s_MemoServ, source, ms_serv_root_help);
	notice(s_MemoServ, source, "\2***** End Of Help *****\2");
    }
}

/*************************************************************************/

/* ms_mmemo()
 * Set the maximum number of memos a user can have, currently this is only affected
 * in a users inbox, as mail is a system call, and trash is 'deleted' memos. */

static void ms_mmemo(const char *source)
{
   char *nick = strtok(NULL, " ");
   char *numstr = strtok(NULL, " ");
   NickInfo *ni;
   User *u;
   int num;

   if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
   }
   if (!nick || !numstr || (num = atoi (numstr)) < 0) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "MMEMO <nick> <number>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP MMEMO");
   } else if (!(ni = findnick(nick))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
   } else {
	ni->memomax = atoi(numstr);
      if (ni->memomax == 0)
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_NOW_HAVE_MEMO_NOLIMIT), nick);
      else
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_NOW_HAVE_MEMO_LIMIT), nick, ni->memomax);
   }
}

/*************************************************************************/

/* ms_send()
 * Send a memo to a nick. */

static void ms_send(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *nick = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    User *u;
    time_t now = time(NULL);

    crash_id = 74;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "SEND <nick> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP SEND");
    } else if (!(ni = findnick(nick))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC), conf.mssnddelay);
    } else if ((ni->flags & NI_NOMEMOS)) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
    } else {
	u->lastmemosend = now;
	ml = find_memolist(nick);
	if (!ml) {
	    ml = scalloc(sizeof(MemoList), 1);
	    strscpy(ml->nick, nick, NICKMAX);
	    alpha_insert_memolist(ml);
	}
      if ((ni->memomax > 0) && (ml->n_memos + 1 > ni->memomax)) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAS_TOO_MANY_MEMOS), ni->nick, ni->memomax);
		return;
	}
	ml->n_memos++;
	ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
	m = &ml->memos[ml->n_memos-1];
	strscpy(m->sender, source, NICKMAX);
	if (ml->n_memos > 1) {
	    m->number = m[-1].number + 1;
	    if (m->number < 1) {
		int i;
		for (i = 0; i < ml->n_memos; ++i)
		    ml->memos[i].number = i+1;
	    }
	} else
	    ml->memos[ml->n_memos-1].number = 1;
	m->time = time(NULL);
	m->text = sstrdup(text);
      m->flags |= (MF_UNREAD | MF_T_NORM | MF_L_INBOX);
      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
	if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(nick))) {
	    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_HAVE_NEW_MEMO_INBOX), source);
	    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
	}
    }
}

/*************************************************************************/

/* ms_reply()
 * Reply to a memo. */

static void ms_reply(const char *source)
{
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    User *u;
    int num, i;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!folder || !text || !numstr || (num = atoi (numstr)) <= 0) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "REPLY [INBOX|TRASH] <number> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP REPLY");
    } else if (!(ni = findnick(source))) {
      notice(s_MemoServ, source, rpl_numeric_lang(source, CR_YOUR_NICK_NOT_REGISTERED));
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (stricmp(folder, "INBOX") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
		return;
	}
      for (i = 0; i < ml->n_memos; i++)
         if (ml->memos[i].number == num && (ml->memos[i].flags & MF_L_INBOX))
		break;
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_NOT_EXIST_INBOX), num);
	} else {
		m = &ml->memos[i];
	      if (!(ni = findnick(m->sender))) {
      		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), m->sender);
			return;
	      }
	      if ((ni->flags & NI_NOMEMOS)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
			return;
		}
	      if ((ni->memomax > 0) && (ml->n_memos + 1 > ni->memomax)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAS_TOO_MANY_MEMOS), ni->nick, ni->memomax);
			return;
		}
		ml = find_memolist(ni->nick);
		if (!ml) {
		    ml = scalloc(sizeof(MemoList), 1);
		    strscpy(ml->nick, m->sender, NICKMAX);
		    alpha_insert_memolist(ml);
		}
		ml->n_memos++;
		ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
		m = &ml->memos[ml->n_memos-1];
		strscpy(m->sender, source, NICKMAX);
		if (ml->n_memos > 1) {
		    m->number = m[-1].number + 1;
		    if (m->number < 1) {
			int i;
			for (i = 0; i < ml->n_memos; ++i)
			    ml->memos[i].number = i+1;
		    }
		} else
		    ml->memos[ml->n_memos-1].number = 1;
		m->time = time(NULL);
		m->text = sstrdup(text);
	      m->flags |= (MF_UNREAD | MF_L_INBOX | MF_T_REPLY);
	      m->flags &= ~(MF_T_FORWARD | MF_T_OPER | MF_T_ADMIN | MF_T_SYS | MF_T_NORM);
	      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
		if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(ni->nick))) {
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_HAVE_NEW_MEMO_INBOX), source);
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		}
	  }
    } else if (stricmp(folder, "TRASH") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
		return;
	}
      for (i = 0; i < ml->n_trash; i++)
         if (ml->trash[i].number == num && (ml->trash[i].flags & MF_L_TRASH))
		break;
	if (i >= ml->n_trash) {
	    notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_NOT_EXIST_TRASH), num);
	} else {
		m = &ml->trash[i];
	      if (!(ni = findnick(m->sender))) {
      		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), m->sender);
			return;
	      }
	      if ((ni->flags & NI_NOMEMOS)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
			return;
		}
		ml = find_memolist(ni->nick);
		if (!ml) {
		    ml = scalloc(sizeof(MemoList), 1);
		    strscpy(ml->nick, m->sender, NICKMAX);
		    alpha_insert_memolist(ml);
		}
		ml->n_memos++;
		ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
		m = &ml->memos[ml->n_memos-1];
		strscpy(m->sender, source, NICKMAX);
		if (ml->n_memos > 1) {
		    m->number = m[-1].number + 1;
		    if (m->number < 1) {
			int i;
			for (i = 0; i < ml->n_memos; ++i)
			    ml->memos[i].number = i+1;
		    }
		} else
		    ml->memos[ml->n_memos-1].number = 1;
		m->time = time(NULL);
		m->text = sstrdup(text);
	      m->flags |= (MF_UNREAD | MF_L_INBOX | MF_T_REPLY);
	      m->flags &= ~(MF_T_FORWARD | MF_T_OPER | MF_T_ADMIN | MF_T_SYS | MF_T_NORM);
	      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
		if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(ni->nick))) {
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_HAVE_NEW_MEMO_INBOX), source);
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		}
	  }
    } else {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "REPLY [INBOX|TRASH] <number> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP REPLY");
    }
}

/*************************************************************************/

/* ms_forward()
 * Forward a memo to another nick. */

static void ms_forward(const char *source)
{
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    MemoList *ml;
    Memo *m, *mm;
    User *u;
    int num, i;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!folder || !nick || !numstr || (num = atoi (numstr)) <= 0) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "FORWARD [INBOX|TRASH] <number> <nick>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP FORWARD");
    } else if (!(ni = findnick(source))) {
      notice(s_MemoServ, source, rpl_numeric_lang(source, CR_YOUR_NICK_NOT_REGISTERED));
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (stricmp(folder, "INBOX") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
		return;
	}
      for (i = 0; i < ml->n_memos; i++)
         if (ml->memos[i].number == num && (ml->memos[i].flags & MF_L_INBOX))
		break;
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_NOT_EXIST_INBOX), num);
	} else {
		mm = &ml->memos[i];
	      if (!(ni = findnick(nick))) {
      		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
	      }
	      if ((ni->flags & NI_NOMEMOS)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
			return;
		}
	      if ((ni->memomax > 0) && (ml->n_memos + 1 > ni->memomax)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAS_TOO_MANY_MEMOS), ni->nick, ni->memomax);
			return;
		}
		ml = find_memolist(ni->nick);
		if (!ml) {
		    ml = scalloc(sizeof(MemoList), 1);
		    strscpy(ml->nick, ni->nick, NICKMAX);
		    alpha_insert_memolist(ml);
		}
		ml->n_memos++;
		ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
		m = &ml->memos[ml->n_memos-1];
		strscpy(m->sender, source, NICKMAX);
		if (ml->n_memos > 1) {
		    m->number = m[-1].number + 1;
		    if (m->number < 1) {
			int i;
			for (i = 0; i < ml->n_memos; ++i)
			    ml->memos[i].number = i+1;
		    }
		} else
		    ml->memos[ml->n_memos-1].number = 1;
		m->time = time(NULL);
		m->text = sstrdup(mm->text);
	      m->flags |= (MF_UNREAD | MF_L_INBOX | MF_T_FORWARD);
	      m->flags &= ~(MF_T_REPLY | MF_T_OPER | MF_T_ADMIN | MF_T_SYS | MF_T_NORM);
	      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
		if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(ni->nick))) {
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_HAVE_NEW_MEMO_INBOX), source);
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		}
	  }
    } else if (stricmp(folder, "TRASH") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
		return;
	}
      for (i = 0; i < ml->n_trash; i++)
         if (ml->trash[i].number == num && (ml->trash[i].flags & MF_L_TRASH))
		break;
	if (i >= ml->n_trash) {
	    notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_NOT_EXIST_TRASH), num);
	} else {
		mm = &ml->trash[i];
	      if (!(ni = findnick(nick))) {
      		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
	      }
	      if ((ni->flags & NI_NOMEMOS)) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
			return;
		}
		ml = find_memolist(ni->nick);
		if (!ml) {
		    ml = scalloc(sizeof(MemoList), 1);
		    strscpy(ml->nick, ni->nick, NICKMAX);
		    alpha_insert_memolist(ml);
		}
		ml->n_memos++;
		ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
		m = &ml->memos[ml->n_memos-1];
		strscpy(m->sender, source, NICKMAX);
		if (ml->n_memos > 1) {
		    m->number = m[-1].number + 1;
		    if (m->number < 1) {
			int i;
			for (i = 0; i < ml->n_memos; ++i)
			    ml->memos[i].number = i+1;
		    }
		} else
		    ml->memos[ml->n_memos-1].number = 1;
		m->time = time(NULL);
		m->text = sstrdup(mm->text);
	      m->flags |= (MF_UNREAD | MF_L_INBOX | MF_T_FORWARD);
	      m->flags &= ~(MF_T_REPLY | MF_T_OPER | MF_T_ADMIN | MF_T_SYS | MF_T_NORM);
	      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
		if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(ni->nick))) {
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_HAVE_NEW_MEMO_INBOX), source);
		    notice(s_MemoServ, u->nick, rpl_numeric_lang(u->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		}
	  }
    } else {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "FORWARD [INBOX|TRASH] <number> <nick>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP FORWARD");
    }
}

/*************************************************************************/

/* ms_send()
 * Send a memo to a nick. */

static void ms_msend(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *nick = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    User *u;
    time_t now = time(NULL);

    crash_id = 74;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "SEND <nick> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP SEND");
    } else if (!(ni = findnick(nick))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC), conf.mssnddelay);
    } else if ((ni->flags & NI_NOMEMOS)) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_DOES_NOT_WANT_MEMOS), ni->nick);
    } else {
	u->lastmemosend = now;
	ml = find_memolist(nick);
	if (!ml) {
	    ml = scalloc(sizeof(MemoList), 1);
	    strscpy(ml->nick, nick, NICKMAX);
	    alpha_insert_memolist(ml);
	}
	ml->n_mail++;
	ml->mail = srealloc(ml->mail, sizeof(Memo) * ml->n_mail);
	m = &ml->mail[ml->n_mail-1];
	strscpy(m->sender, source, NICKMAX);
	if (ml->n_mail > 1) {
	    m->number = m[-1].number + 1;
	    if (m->number < 1) {
		int i;
		for (i = 0; i < ml->n_mail; ++i)
		    ml->mail[i].number = i+1;
	    }
	} else
	    ml->mail[ml->n_mail-1].number = 1;
	m->time = time(NULL);
	m->text = sstrdup(text);
      m->flags |= (MF_UNREAD | MF_T_MAIL | MF_L_MAIL);
      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ni->nick);
	if ((ni->flags & NI_MEMO_RECEIVE) && finduser(nick)) {
	    notice(s_MemoServ, nick, rpl_numeric_lang(nick, MS_HAVE_NEW_MEMO_MAIL), source);
	    notice(s_MemoServ, nick, rpl_numeric_lang(nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "MAIL", m->number);
	}
    }
}

/*************************************************************************/

/* ms_list()
 * List the memos (if any) for the source nick from the target folder, or all folders. */

static void ms_list(const char *source)
{
    char *folder = strtok(NULL, " ");
    char *param = strtok(NULL, " ");
    MemoList *ml;
    Memo *m;
    int i, inbox_cnt = 0, mail_cnt = 0, trash_cnt = 0, show_new = 0;
    char timebuf[64];
    struct tm tm;

    crash_id = 75;

    if (param && stricmp(param, "NEW") == 0) show_new = 1;

    if (folder) {
	if (stricmp(folder, "INBOX") == 0) {
		if (!(ml = find_memolist(source))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
			return;
		}
		if (show_new) {
		    for (i = 0, m = ml->memos; i < ml->n_memos; i++, m++) {
			if ((m->flags & MF_UNREAD) && (m->flags & MF_L_INBOX))
			    break;
		    }
		    if (i == ml->n_memos) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS_INBOX));
			return;
		    }
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_INBOX_N), source, s_MemoServ);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_INBOX), source, s_MemoServ);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) {
		    if (m->flags & MF_L_INBOX) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, (m->flags & MF_UNREAD) ? 'U' : 'R', strlen(m->text), m->sender, timebuf);
			}
	 	} 
    	} else if (stricmp(folder, "MAIL") == 0) {
		if (!(ml = find_memolist(source))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
			return;
		}
		if (show_new) {
		    for (i = 0, m = ml->mail; i < ml->n_mail; i++, m++) {
			if ((m->flags & MF_UNREAD) && (m->flags & MF_L_MAIL))
			    break;
		    }
		    if (i == ml->n_mail) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS_MAIL));
			return;
		    }
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_MAIL_N), source, s_MemoServ);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_MAIL), source, s_MemoServ);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->mail; i < ml->n_mail; ++i, ++m) {
		    if (m->flags & MF_L_MAIL) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, (m->flags & MF_UNREAD) ? 'U' : 'R', strlen(m->text), m->sender, timebuf);
			}
	 	} 
    	} else if (stricmp(folder, "TRASH") == 0) {
		if (!(ml = find_memolist(source))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
			return;
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_TRASH_N), source, s_MemoServ);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_TRASH), source, s_MemoServ);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->trash; i < ml->n_trash; ++i, ++m) {
		    if (m->flags & MF_L_TRASH) {
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
			    /* There are no new memos in this folder, as they are trash and we do not monitor it, simply
			     * list the memos, and be done.  Status - D relates to Deleted. */
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, 'D', strlen(m->text), m->sender, timebuf);
			}
	 	} 
	} else {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "LIST [\37INBOX\37|\37MAIL\37|\37TRASH\37] [\37NEW\37]");
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP LIST");
	}
    } else {
	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_YOU_HAVE_NO_MEMOS));
	} else {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_HEADER_FOLDERS), source, s_MemoServ);
		for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) if (m->flags & MF_L_INBOX) inbox_cnt++;
		for (i = 0, m = ml->mail; i < ml->n_mail; ++i, ++m) if (m->flags & MF_L_MAIL) mail_cnt++;
		for (i = 0, m = ml->trash; i < ml->n_trash; ++i, ++m) if (m->flags & MF_L_TRASH) trash_cnt++;
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FOLDERS), inbox_cnt, mail_cnt, trash_cnt);
	}
   }
}

/*************************************************************************/

/* ms_rlist()
 * Remote list the memos folders of another nick other then you own. */

static void ms_rlist(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *param = strtok(NULL, " ");
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    int i, inbox_cnt = 0, mail_cnt = 0, trash_cnt = 0, show_new = 0;
    char timebuf[64];
    struct tm tm;

    crash_id = 75;

    if (param && stricmp(param, "NEW") == 0) show_new = 1;

    if (nick) {
	if (folder && stricmp(folder, "INBOX") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
		}
		if (!(ml = find_memolist(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS), nick);
			return;
		}
		if (show_new) {
		    for (i = 0, m = ml->memos; i < ml->n_memos; i++, m++) {
			if ((m->flags & MF_UNREAD) && (m->flags & MF_L_INBOX))
			    break;
		    }
		    if (i == ml->n_memos) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS_INBOX), nick);
			return;
		    }
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_INBOX_N), ni->nick, s_MemoServ, ni->nick);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_INBOX), ni->nick, s_MemoServ, ni->nick);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) {
		    if (m->flags & MF_L_INBOX) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, (m->flags & MF_UNREAD) ? 'U' : 'R', strlen(m->text), m->sender, timebuf);
			}
	 	} 
    	} else if (folder && stricmp(folder, "MAIL") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
		}
		if (!(ml = find_memolist(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS), nick);
			return;
		}
		if (show_new) {
		    for (i = 0, m = ml->mail; i < ml->n_mail; i++, m++) {
			if ((m->flags & MF_UNREAD) && (m->flags & MF_L_MAIL))
			    break;
		    }
		    if (i == ml->n_mail) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS_MAIL), nick);
			return;
		    }
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_MAIL_N), ni->nick, s_MemoServ, ni->nick);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_MAIL), ni->nick, s_MemoServ, ni->nick);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->mail; i < ml->n_mail; ++i, ++m) {
		    if (m->flags & MF_L_MAIL) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, (m->flags & MF_UNREAD) ? 'U' : 'R', strlen(m->text), m->sender, timebuf);
			}
	 	} 
    	} else if (folder && stricmp(folder, "TRASH") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
		}
		if (!(ml = find_memolist(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS), nick);
			return;
		}
		if (param)
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_TRASH_N), ni->nick, s_MemoServ, ni->nick);
		else
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_TRASH), ni->nick, s_MemoServ, ni->nick);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = ml->trash; i < ml->n_trash; ++i, ++m) {
		    if (m->flags & MF_L_TRASH) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FORMAT),
				   m->number, 'D', strlen(m->text), m->sender, timebuf);
			}
	 	} 
	} else {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
			return;
		}
		if (!(ml = find_memolist(nick))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_HAVE_NO_MEMOS), nick);
		} else {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_NICK_MEMO_LIST_HEADER_FOLDERS), ni->nick, s_MemoServ, ni->nick);
			for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) if (m->flags & MF_L_INBOX) inbox_cnt++;
			for (i = 0, m = ml->mail; i < ml->n_mail; ++i, ++m) if (m->flags & MF_L_MAIL) mail_cnt++;
			for (i = 0, m = ml->trash; i < ml->n_trash; ++i, ++m) if (m->flags & MF_L_TRASH) trash_cnt++;
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LIST_FOLDERS), inbox_cnt, mail_cnt, trash_cnt);
		}
	}
    } else {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "RLIST <nick> [\37INBOX\37|\37MAIL\37|\37TRASH\37] [\37NEW\37]");
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP RLIST");
   }
}

/*************************************************************************/

/* ms_read()
 * Read a memo in the target folder. */

static void ms_read(const char *source)
{
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    MemoList *ml;
    Memo *m;
    int num, i, show_last = 0;

    crash_id = 76;

    if (numstr && stricmp(numstr, "LAST") == 0) show_last = 1;
 
    if (!folder || (numstr && stricmp(numstr, "LAST") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: READ [INBOX|MAIL|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP READ\2.", s_MemoServ);
    } else if (stricmp(folder, "INBOX") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_memos-1 && (ml->memos[i].flags & MF_L_INBOX); i++)
		;
	} else {
	    for (i = 0; i < ml->n_memos; i++) {
		if (ml->memos[i].number == num && (ml->memos[i].flags & MF_L_INBOX))
		    break;
	    }
	}
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d in your inbox folder does not exist.", num);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in your inbox folder from %s (%s).  To delete this memo in your inbox folder, please type \2/msg %s DEL INBOX %d\2.",
		m->number, m->sender, timebuf, s_MemoServ, m->number);
	    if ((m->flags & MF_T_SYS))
		    notice(s_MemoServ, source, "Memo type: System");
	    else if ((m->flags & MF_T_ADMIN))
		    notice(s_MemoServ, source, "Memo type: Administrator");
	    else if ((m->flags & MF_T_OPER))
		    notice(s_MemoServ, source, "Memo type: IRC Operator");
	    else if ((m->flags & MF_T_NORM))
		    notice(s_MemoServ, source, "Memo type: User");
	    else if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    else if ((m->flags & MF_T_REPLY))
		    notice(s_MemoServ, source, "Memo type: Replied");
	    else if ((m->flags & MF_T_FORWARD))
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    else
		    notice(s_MemoServ, source, "Memo type: User");
	    notice(s_MemoServ, source, "%s", m->text);
	    m->flags &= ~MF_UNREAD;
	}
    } else if (stricmp(folder, "MAIL") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_mail-1 && (ml->mail[i].flags & MF_L_MAIL); i++)
		;
	} else {
	    for (i = 0; i < ml->n_mail; i++) {
		if (ml->mail[i].number == num && (ml->mail[i].flags & MF_L_MAIL))
		    break;
	    }
	}
	if (i >= ml->n_mail) {
	    notice(s_MemoServ, source, "Memo %d in your mail folder does not exist.", num);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->mail[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in your mail folder from %s (%s).  To delete this memo in your mail folder, please type \2/msg %s DEL MAIL %d\2.",
		m->number, m->sender, timebuf, s_MemoServ, m->number);
	    if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    notice(s_MemoServ, source, "%s", m->text);
	    m->flags &= ~MF_UNREAD;
	}
    } else if (stricmp(folder, "TRASH") == 0) {
    	if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_trash-1 && (ml->trash[i].flags & MF_L_TRASH); i++)
		;
	} else {
	    for (i = 0; i < ml->n_trash; i++) {
		if (ml->trash[i].number == num && (ml->trash[i].flags & MF_L_TRASH))
		    break;
	    }
	}
	if (i >= ml->n_trash) {
	    notice(s_MemoServ, source, "Memo %d in your trash folder does not exist.", num);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->trash[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in your trash folder from %s (%s).  To delete this memo in your trash folder, please type \2/msg %s DEL TRASH %d\2 or to undelete this memo type \2/msg %s UNDEL %d\2",
		m->number, m->sender, timebuf, s_MemoServ, m->number, s_MemoServ, m->number);
	    if ((m->flags & MF_T_SYS))
		    notice(s_MemoServ, source, "Memo type: System");
	    else if ((m->flags & MF_T_ADMIN))
		    notice(s_MemoServ, source, "Memo type: Administrator");
	    else if ((m->flags & MF_T_OPER))
		    notice(s_MemoServ, source, "Memo type: IRC Operator");
	    else if ((m->flags & MF_T_NORM))
		    notice(s_MemoServ, source, "Memo type: User");
	    else if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    else if ((m->flags & MF_T_REPLY))
		    notice(s_MemoServ, source, "Memo type: Replied");
	    else if ((m->flags & MF_T_FORWARD))
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    else
		    notice(s_MemoServ, source, "Memo type: User");

	    /* Since there is only two locations, anything else is unknown; there must have been a problem. */
	    if ((m->flags & MF_L_INBOX))
	    	    notice(s_MemoServ, source, " Location: Previously stored in your inbox folder.");
	    else if ((m->flags & MF_L_MAIL))
	    	    notice(s_MemoServ, source, " Location: Previously stored in your mail folder.");
	    else
	    	    notice(s_MemoServ, source, " Location: The folder this memo was previously stored in is unknown.");
	    notice(s_MemoServ, source, "%s", m->text);
	    m->flags &= ~MF_UNREAD;
	}
    } else {
	notice(s_MemoServ, source, "Syntax: READ [INBOX|MAIL|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP READ\2.", s_MemoServ);
    }
}

/*************************************************************************/

/* ms_rread()
 * Remote read memos in the memo folders of another nick other then you own. */

static void ms_rread(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    int num, i, show_last = 0;

    crash_id = 76;

    if (numstr && stricmp(numstr, "LAST") == 0) show_last = 1;
 
    if (!nick || !folder || (numstr && stricmp(numstr, "LAST") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: RREAD <nick> [INBOX|MAIL|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP RREAD\2.", s_MemoServ);
    } else if (stricmp(folder, "INBOX") == 0) {
	if (!(ni = findnick(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	}
    	if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.", nick);
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_memos-1 && (ml->memos[i].flags & MF_L_INBOX); i++)
		;
	} else {
	    for (i = 0; i < ml->n_memos; i++) {
		if (ml->memos[i].number == num && (ml->memos[i].flags & MF_L_INBOX))
		    break;
	    }
	}
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d in \2%s\2 inbox folder does not exist.", num, nick);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in %s inbox folder from %s (%s).  To delete this memo in thier inbox folder, please type \2/msg %s RDEL %s INBOX %d\2.",
		m->number, ni->nick, m->sender, timebuf, s_MemoServ, ni->nick, m->number);
	    if ((m->flags & MF_T_SYS))
		    notice(s_MemoServ, source, "Memo type: System");
	    else if ((m->flags & MF_T_ADMIN))
		    notice(s_MemoServ, source, "Memo type: Administrator");
	    else if ((m->flags & MF_T_OPER))
		    notice(s_MemoServ, source, "Memo type: IRC Operator");
	    else if ((m->flags & MF_T_NORM))
		    notice(s_MemoServ, source, "Memo type: User");
	    else if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    else if ((m->flags & MF_T_REPLY))
		    notice(s_MemoServ, source, "Memo type: Replied");
	    else if ((m->flags & MF_T_FORWARD))
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    else
		    notice(s_MemoServ, source, "Memo type: User");
	    notice(s_MemoServ, source, "%s", m->text);
	}
    } else if (stricmp(folder, "MAIL") == 0) {
	if (!(ni = findnick(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	}
    	if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.");
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_mail-1 && (ml->mail[i].flags & MF_L_MAIL); i++)
		;
	} else {
	    for (i = 0; i < ml->n_mail; i++) {
		if (ml->mail[i].number == num && (ml->mail[i].flags & MF_L_MAIL))
		    break;
	    }
	}
	if (i >= ml->n_mail) {
	    notice(s_MemoServ, source, "Memo %d in \2%s\2 mail folder does not exist.", num, nick);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->mail[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in %s mail folder from %s (%s).  To delete this memo in your mail folder, please type \2/msg %s RDEL %s MAIL %d\2.",
		m->number, ni->nick, m->sender, timebuf, s_MemoServ, ni->nick, m->number);
	    if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    notice(s_MemoServ, source, "%s", m->text);
	}
    } else if (stricmp(folder, "TRASH") == 0) {
	if (!(ni = findnick(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	}
    	if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.");
		return;
	}
	if (show_last) {
	    for (i = 0; i < ml->n_trash-1 && (ml->trash[i].flags & MF_L_TRASH); i++)
		;
	} else {
	    for (i = 0; i < ml->n_trash; i++) {
		if (ml->trash[i].number == num && (ml->trash[i].flags & MF_L_TRASH))
		    break;
	    }
	}
	if (i >= ml->n_trash) {
	    notice(s_MemoServ, source, "Memo %d in \2%s\2 trash folder does not exist.", num, nick);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &ml->trash[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in %s trash folder from %s (%s).  To delete this memo in your trash folder, please type \2/msg %s RDEL %s TRASH %d\2.",
		m->number, ni->nick, m->sender, timebuf, s_MemoServ, ni->nick, m->number);
	    if ((m->flags & MF_T_SYS))
		    notice(s_MemoServ, source, "Memo type: System");
	    else if ((m->flags & MF_T_ADMIN))
		    notice(s_MemoServ, source, "Memo type: Administrator");
	    else if ((m->flags & MF_T_OPER))
		    notice(s_MemoServ, source, "Memo type: IRC Operator");
	    else if ((m->flags & MF_T_NORM))
		    notice(s_MemoServ, source, "Memo type: User");
	    else if ((m->flags & MF_T_MAIL))
		    notice(s_MemoServ, source, "Memo type: Mail");
	    else if ((m->flags & MF_T_REPLY))
		    notice(s_MemoServ, source, "Memo type: Replied");
	    else if ((m->flags & MF_T_FORWARD))
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    else
		    notice(s_MemoServ, source, "Memo type: User");
	    notice(s_MemoServ, source, "%s", m->text);
	}
    } else {
	notice(s_MemoServ, source, "Syntax: RREAD <nick> [INBOX|MAIL|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP RREAD\2.", s_MemoServ);
    }
}

/*************************************************************************/

/* ms_del()
 * Delete a memo. */

static void ms_del(const char *source)
{
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    MemoList *ml;
    Memo *m;
    int num, i, del_all = 0;

    crash_id = 77;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }

    if (numstr && stricmp(numstr, "ALL") == 0) del_all = 1;
 
    if (!folder || (numstr && stricmp(numstr, "ALL") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: DEL [INBOX|MAIL|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP DEL\2.", s_MemoServ);
    } else if (stricmp(folder, "INBOX") == 0) {
	    if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	    }
          if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_memos; ++i) {
			if (ml->memos[i].number == num)
			    break;
		    }
			/* Send this memo, to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->memos[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->memos[i].text);
		      m->flags = ml->memos[i].flags;
			/* We keep this flag for notices in a reading of a trash memo and for where to undelete it to. */
			m->flags &= ~(MF_UNREAD);
			// m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags = MF_L_TRASH;

		    if (i < ml->n_memos) {
			free(ml->memos[i].text); 		/* Deallocate memo text memory */
			ml->n_memos--;		 		/* One less memo now */
			if (i < ml->n_memos)	 		/* Move remaining memos down a slot */
			    bcopy(ml->memos + i+1, ml->memos + i, sizeof(Memo) * (ml->n_memos - i));
			notice(s_MemoServ, source, "Memo %d in your inbox folder has been deleted.", num);
		    } else {
			notice(s_MemoServ, source, "Memo %d in your inbox folder does not exist.", num);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_memos; ++i) {
			/* Send this memo, to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->memos[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->memos[i].text);
		      m->flags = ml->memos[i].flags;
			m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags = MF_L_TRASH;

			free(ml->memos[i].text);
		    }
		    ml->n_memos = 0;
		    notice(s_MemoServ, source, "All of your memos in your inbox folder have been deleted.");
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else if (stricmp(folder, "MAIL") == 0) {
	    if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	    }
		if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_mail; ++i) {
			if (ml->mail[i].number == num)
			    break;
		    }
			/* Send this memo, to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->mail[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->mail[i].text);
		      m->flags = ml->mail[i].flags;
			/* We keep this flag for notices in a reading of a trash memo and for where to undelete it to. */
			m->flags &= ~(MF_UNREAD);
			// m->flags &= ~(MF_L_MAIL | MF_UNREAD);
			m->flags = MF_L_TRASH;

		    if (i < ml->n_mail) {
			free(ml->mail[i].text); 		/* Deallocate memo text memory */
			ml->n_mail--;		 		/* One less memo now */
			if (i < ml->n_mail)	 		/* Move remaining memos down a slot */
			    bcopy(ml->mail + i+1, ml->mail + i, sizeof(Memo) * (ml->n_mail - i));
			notice(s_MemoServ, source, "Memo %d in your mail folder has been deleted.", num);
		    } else {
			notice(s_MemoServ, source, "Memo %d in your mail folder does not exist.", num);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_mail; ++i) {
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->mail[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->mail[i].text);
		      m->flags = ml->mail[i].flags;
			m->flags &= ~(MF_L_MAIL | MF_UNREAD);
			m->flags = MF_L_TRASH;

			free(ml->mail[i].text);
		    }
		    ml->n_mail = 0;
		    notice(s_MemoServ, source, "All of your memos in your mail folder have been deleted.");
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else if (stricmp(folder, "TRASH") == 0) {
	    if (!(ml = find_memolist(source))) {
		notice(s_MemoServ, source, "You have no memos.");
		return;
	    }
		if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_trash; ++i) {
			if (ml->trash[i].number == num)
			    break;
		    }
		    if (i < ml->n_trash) {
			free(ml->trash[i].text); 		/* Deallocate memo text memory */
			ml->n_trash--;		 		/* One less memo now */
			if (i < ml->n_trash)	 		/* Move remaining memos down a slot */
			    bcopy(ml->trash + i+1, ml->trash + i, sizeof(Memo) * (ml->n_trash - i));
			notice(s_MemoServ, source, "Memo %d in your trash folder has been deleted.", num);
		    } else {
			notice(s_MemoServ, source, "Memo %d in your trash folder does not exist.", num);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_trash; ++i)
			free(ml->trash[i].text);
		    ml->n_trash = 0;
		    notice(s_MemoServ, source, "All of your memos in your trash folder have been deleted.");
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else {
	notice(s_MemoServ, source, "Syntax: DEL [INBOX|MAIL|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP DEL\2.", s_MemoServ);
    }
}

/*************************************************************************/

/* ms_undel()
 * Undelete a memo. */

static void ms_undel(const char *source)
{
    char *numstr = strtok(NULL, " ");
    MemoList *ml;
    Memo *m;
    int num, i;

    crash_id = 77;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }

    if ((num = atoi(numstr)) <= 0) {
	notice(s_MemoServ, source, "Syntax: UNDEL <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP UNDEL\2.", s_MemoServ);
    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");
    } else {
       /* Undelete a specific memo. */	
       for (i = 0; i < ml->n_trash; ++i) {
	   if (ml->trash[i].number == num)
	      break;
       }
 	    /* Return this memo to where it was before. */
	    if ((ml->trash[i].flags & MF_L_INBOX)) {
		if (num > 0) {
			ml->n_memos++;
			ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos-1];
			strscpy(m->sender, ml->trash[i].sender, NICKMAX);
			if (ml->n_memos > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_memos; ++j)
				    ml->memos[j].number = j+1;
			    }
			} else
			    ml->memos[ml->n_memos-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->trash[i].text);
			m->flags = ml->trash[i].flags;
			/* We keep this flag for notices in a reading of a trash memo and for where to undelete it to. */
			m->flags &= ~(MF_L_TRASH | MF_UNREAD);
			m->flags = MF_L_INBOX;
         
    	            if (i < ml->n_trash) {
			   free(ml->trash[i].text); 		/* Deallocate memo text memory */
			   ml->n_trash--;		 		/* One less memo now */
			   if (i < ml->n_trash)	 		/* Move remaining memos down a slot */
				bcopy(ml->trash + i+1, ml->trash + i, sizeof(Memo) * (ml->n_trash - i));
			   notice(s_MemoServ, source, "Memo %d in your trash folder has been undeleted and restored to your inbox folder.", num);
		        } else {
			   notice(s_MemoServ, source, "Memo %d in your trash folder does not exist.", num);
		        }
		  }
	     }

	    if ((ml->trash[i].flags & MF_L_MAIL)) {
		if (num > 0) {
			ml->n_mail++;
			ml->mail = srealloc(ml->mail, sizeof(Memo) * ml->n_mail);
			m = &ml->mail[ml->n_mail-1];
			strscpy(m->sender, ml->trash[i].sender, NICKMAX);
			if (ml->n_mail > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_mail; ++j)
				    ml->mail[j].number = j+1;
			    }
			} else
			    ml->mail[ml->n_mail-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->trash[i].text);
			m->flags = ml->trash[i].flags;
			/* We keep this flag for notices in a reading of a trash memo and for where to undelete it to. */
			m->flags &= ~(MF_L_TRASH | MF_UNREAD);
			m->flags = MF_L_MAIL;
	         
	                if (i < ml->n_trash) {
			   free(ml->trash[i].text); 		/* Deallocate memo text memory */
			   ml->n_trash--;		 		/* One less memo now */
			   if (i < ml->n_trash)	 		/* Move remaining memos down a slot */
				bcopy(ml->trash + i+1, ml->trash + i, sizeof(Memo) * (ml->n_trash - i));
			   notice(s_MemoServ, source, "Memo %d in your trash folder has been undeleted and restored to your inbox folder.", num);
		        } else {
			   notice(s_MemoServ, source, "Memo %d in your trash folder does not exist.", num);
		        }
		  }
	     }
     }
}

/*************************************************************************/

/* ms_rdel()
 * Remote delete of another nick's memos. */

static void ms_rdel(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    int num, i, del_all = 0;

    crash_id = 77;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }

    if (numstr && stricmp(numstr, "ALL") == 0) del_all = 1;
 
    if (!nick || !folder || (numstr && stricmp(numstr, "ALL") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: RDEL <nick> [INBOX|MAIL|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP RDEL\2.", s_MemoServ);
    } else if (stricmp(folder, "INBOX") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
			return;
		}
	    if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.", nick);
		return;
	    }
          if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_memos; ++i) {
			if (ml->memos[i].number == num)
			    break;
		    }
			/* Move this memo to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->memos[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->memos[i].text);
		      m->flags = ml->memos[i].flags;
			m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags = MF_L_TRASH;

		    if (i < ml->n_memos) {
			free(ml->memos[i].text); 		/* Deallocate memo text memory */
			ml->n_memos--;		 		/* One less memo now */
			if (i < ml->n_memos)	 		/* Move remaining memos down a slot */
			    bcopy(ml->memos + i+1, ml->memos + i, sizeof(Memo) * (ml->n_memos - i));
			notice(s_MemoServ, source, "Memo %d in \2%s\2 inbox folder has been deleted.", num, nick);
		    } else {
			notice(s_MemoServ, source, "Memo %d in \2%s\2 inbox folder does not exist.", num, nick);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_memos; ++i) {
			/* Send this memo, to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->memos[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->memos[i].text);
		      m->flags = ml->memos[i].flags;
			m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags = MF_L_TRASH;

			free(ml->memos[i].text);
		    }
		    ml->n_memos = 0;
		    notice(s_MemoServ, source, "All of \2%s\2 memos in their inbox folder have been deleted.", nick);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else if (stricmp(folder, "MAIL") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
			return;
		}
	    if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.", nick);
		return;
	    }
		if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_mail; ++i) {
			if (ml->mail[i].number == num)
			    break;
		    }
			/* Send this memo, to the trash folder. */
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->mail[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->mail[i].text);
		      m->flags = ml->mail[i].flags;
			m->flags &= ~(MF_L_MAIL | MF_UNREAD);
			m->flags = MF_L_TRASH;

		    if (i < ml->n_mail) {
			free(ml->mail[i].text); 		/* Deallocate memo text memory */
			ml->n_mail--;		 		/* One less memo now */
			if (i < ml->n_mail)	 		/* Move remaining memos down a slot */
			    bcopy(ml->mail + i+1, ml->mail + i, sizeof(Memo) * (ml->n_mail - i));
			notice(s_MemoServ, source, "Memo %d in \2%s\2 mail folder has been deleted.", num, nick);
		    } else {
			notice(s_MemoServ, source, "Memo %d in \2%s\2 mail folder does not exist.", num, nick);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_mail; ++i) {
			ml->n_trash++;
			ml->trash = srealloc(ml->trash, sizeof(Memo) * ml->n_trash);
			m = &ml->trash[ml->n_trash-1];
			strscpy(m->sender, ml->mail[i].sender, NICKMAX);
			if (ml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < ml->n_trash; ++j)
				    ml->trash[j].number = j+1;
			    }
			} else
			    ml->trash[ml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(ml->mail[i].text);
		      m->flags = ml->mail[i].flags;
			m->flags &= ~(MF_L_MAIL | MF_UNREAD);
			m->flags = MF_L_TRASH;

			free(ml->mail[i].text);
		    }
		    ml->n_mail = 0;
		    notice(s_MemoServ, source, "All of \2%s\2 memos in their mail folder have been deleted.", nick);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else if (stricmp(folder, "TRASH") == 0) {
		if (!(ni = findnick(nick))) {
			notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
			return;
		}
	    if (!(ml = find_memolist(nick))) {
		notice(s_MemoServ, source, "Nick \2%s\2 has no memos.", nick);
		return;
	    }
		if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < ml->n_trash; ++i) {
			if (ml->trash[i].number == num)
			    break;
		    }
		    if (i < ml->n_trash) {
			free(ml->trash[i].text); 		/* Deallocate memo text memory */
			ml->n_trash--;		 		/* One less memo now */
			if (i < ml->n_trash)	 		/* Move remaining memos down a slot */
			    bcopy(ml->trash + i+1, ml->trash + i, sizeof(Memo) * (ml->n_trash - i));
			notice(s_MemoServ, source, "Memo %d in \2%s\2 trash folder has been deleted.", num, nick);
		    } else {
			notice(s_MemoServ, source, "Memo %d in \2%s\2 trash folder does not exist.", num, nick);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < ml->n_trash; ++i)
			free(ml->trash[i].text);
		    ml->n_trash = 0;
		    notice(s_MemoServ, source, "All of \2%s\2 memos in their trash folder have been deleted.", nick);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (ml->n_mail == 0 && ml->n_memos == 0 && ml->n_trash == 0)
		    del_memolist(ml);
    } else {
	notice(s_MemoServ, source, "Syntax: RDEL <nick> [INBOX|MAIL|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP RDEL\2.", s_MemoServ);
    }
}

/*************************************************************************/

/* ms_set()
 * Set various parameters related to memos and how MemoServ notifies users
 * of their memos. */

static void ms_set(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *param = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    User *u;

    crash_id = 78;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!param) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "SET <option> <parameters>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP SET");
    } else if (!(ni = findnick(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_YOUR_NICK_NOT_REGISTERED));
    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PASSWORD_AUTH_REQUIRED));
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_RETRY_AFTER_TYPE), s_NickServ);
    } else if (stricmp(cmd, "NOTIFY") == 0) {
	ms_set_notify(ni, param);
    } else {
	 notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SET_UNKNOWN), cmd, s_MemoServ);
    }
}

/*************************************************************************/

/* ms_set_notify()
 * Set the way MemoServ notifies users of their memos. */

static void ms_set_notify(NickInfo *ni, char *param)
{
    crash_id = 79;

    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_MEMO_SET_NOTIFY_ON), s_MemoServ);
    } else if (stricmp(param, "LOGON") == 0) {
	ni->flags |= NI_MEMO_SIGNON;
	ni->flags &= ~NI_MEMO_RECEIVE;
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_MEMO_SET_NOTIFY_LOGON), s_MemoServ);
    } else if (stricmp(param, "NEW") == 0) {
	ni->flags &= ~NI_MEMO_SIGNON;
	ni->flags |= NI_MEMO_RECEIVE;
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_MEMO_SET_NOTIFY_NEW), s_MemoServ);
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~(NI_MEMO_SIGNON | NI_MEMO_RECEIVE);
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_MEMO_SET_NOTIFY_OFF), s_MemoServ);
    } else {
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, CR_CMD_SYNTAX), "SET NOTIFY [ON|LOGON|NEW|OFF]");
	notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, CR_CMD_MORE_INFO), s_MemoServ, "SET NOTIFY");
    }
}

/*************************************************************************/

/* ms_mail()
 * Send email from IRC to the world. */

static void ms_mail(const char *source)
{
   char *nick = strtok(NULL, " ");
   char *message = strtok(NULL, "");
   NickInfo *sni, *dni;
   FILE *mail;
   char buf[BUFSIZE];
   int i;

   crash_id = 80;

   if (!nick || !message) {
	notice(s_MemoServ, source, "Syntax: MAIL <nick> <message>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP MAIL\2.", s_MemoServ);
   } else if (!(dni = findnick(nick))) {
	notice(s_MemoServ, source, "Nick \2%s\2 is not registered.", nick);
   } else if ((sni = findnick(source)) && !sni->email) {
	notice(s_MemoServ, source, "You do not have your email address set on your nick.");
	notice(s_MemoServ, source, "Please type \2/msg %s SET EMAIL <address>\2 to set it.", s_NickServ);
   } else if (!dni->email) {
	notice(s_MemoServ, source, "\2%s\2 does not have their email address set.", dni->nick);
   } else {
     snprintf(buf, sizeof(buf), "%s %s", conf.msmailbinary, dni->email);
     mail = popen(buf, "w");
     if (!mail) {
        notice(s_MemoServ, source, "Mail for \2%s (%s)\2 could not be sent for delivery.", dni->nick, dni->email);     
     } else {
        /* fprintf(mail, "To: %s <%s>\n", dni->nick, dni->email); */
        fprintf(mail, "From: %s <%s>\n", sni->nick, sni->email);
        fprintf(mail, "Subject: %s\n", conf.msmailsubject);
        fprintf(mail, "%s\n", message);
        fprintf(mail, "\n\n%s\n", conf.msmailfooter);
        notice(s_MemoServ, source, "Mail for \2%s (%s)\2 has been sent for delivery.", dni->nick, dni->email);
        pclose(mail);
     }
   }
}

/*************************************************************************/

/* ms_send_register()
 * Send a memo to a register of a nick. */

void ms_send_register(const char *source, const char *nick)
{
    MemoList *ml;
    Memo *m;
    User *u;
    time_t now = time(NULL);

    crash_id = 81;

    ml = find_memolist(nick);
    if (!ml) {
       ml = scalloc(sizeof(MemoList), 1);
	strscpy(ml->nick, nick, NICKMAX);
	alpha_insert_memolist(ml);
    }
    ml->n_memos++;
    ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
    m = &ml->memos[ml->n_memos-1];
    strscpy(m->sender, source, NICKMAX);
    if (ml->n_memos > 1) {
       m->number = m[-1].number + 1;
	if (m->number < 1) {
	   int i;
	   for (i = 0; i < ml->n_memos; ++i)
	       ml->memos[i].number = i+1;
        }
    } else {
       ml->memos[ml->n_memos-1].number = 1;
    }
    m->time = time(NULL);
    m->text = sstrdup(conf.nsregistermemo);
    m->flags |= (MF_UNREAD | MF_T_SYS | MF_L_INBOX);
    if ((u = finduser(nick))) {
       notice(s_MemoServ, nick, rpl_numeric_lang(nick, MS_HAVE_NEW_MEMO_INBOX), source);
	 notice(s_MemoServ, nick, rpl_numeric_lang(nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
    }
}

/*************************************************************************/

/* ms_opersend()
 * Send a memo to all IRC Operators. */

static void ms_opersend(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *text = strtok(NULL, "");
    User *u;
    time_t now = time(NULL);
    int i;

    crash_id = 82;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "OPERSEND <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP OPERSEND");
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC), conf.mssnddelay);
    } else {
	 u->lastmemosend = now;
       for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if ((ni->flags & NI_IRCOP) && stricmp(ni->nick, source) != 0) {
			ml = find_memolist(ni->nick);
			if (!ml) {
			    ml = scalloc(sizeof(MemoList), 1);
			    strscpy(ml->nick, ni->nick, NICKMAX);
			    alpha_insert_memolist(ml);
			}
			ml->n_memos++;
			ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos-1];
			strscpy(m->sender, source, NICKMAX);
			if (ml->n_memos > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int i;
				for (i = 0; i < ml->n_memos; ++i)
				    ml->memos[i].number = i+1;
			    }
			} else
			    ml->memos[ml->n_memos-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(text);
		      m->flags = MF_UNREAD | MF_T_OPER | MF_L_INBOX;
			if ((ni->flags & NI_MEMO_RECEIVE) && finduser(ni->nick)) {
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_HAVE_NEW_MEMO_INBOX), source);
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		   }
	      }
	   }
      }
	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT_OPER));
   }
}

/*************************************************************************/

/* ms_usersend()
 * Send a memo to all users. */

static void ms_usersend(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *text = strtok(NULL, "");
    User *u;
    time_t now = time(NULL);
    int i;

    crash_id = 83;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "USERSEND <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP USERSEND");
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC), conf.mssnddelay);
    } else {
	 u->lastmemosend = now;
       for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (stricmp(ni->nick, source) != 0) {
			ml = find_memolist(ni->nick);
			if (!ml) {
			    ml = scalloc(sizeof(MemoList), 1);
			    strscpy(ml->nick, ni->nick, NICKMAX);
			    alpha_insert_memolist(ml);
			}
			ml->n_memos++;
			ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos-1];
			strscpy(m->sender, source, NICKMAX);
			if (ml->n_memos > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int i;
				for (i = 0; i < ml->n_memos; ++i)
				    ml->memos[i].number = i+1;
			    }
			} else
			    ml->memos[ml->n_memos-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(text);
		      m->flags = MF_UNREAD | MF_T_SYS | MF_L_INBOX;
			if ((ni->flags & NI_MEMO_RECEIVE) && finduser(ni->nick)) {
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_HAVE_NEW_MEMO_INBOX), source);
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		   }
	      }
	   }
      }
	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT_USER));
   }
}

/*************************************************************************/

/* ms_adminsend()
 * Send a memo to all level x administrators. */

static void ms_adminsend(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *level = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    int lvl, i;
    User *u;
    time_t now = time(NULL);

    crash_id = 84;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "ADMINSEND <level> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP ADMINSEND");
    } else if (!(lvl = atoi(level))) {
      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LEVEL_MUST_BE_INT));
    } else if (lvl < 1 || lvl > 3) {
 	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_LEVEL_MUST_INT_INC), SA_ADMIN, SA_SR);
    } else if (get_admin_level(source) < lvl) {
      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_NO_SEND_ADMIN_HIGH));
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC),
         conf.mssnddelay);
    } else {
	 u->lastmemosend = now;
       for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if ((ni->flags & NI_IRCOP) && is_serv_admin(ni->nick, lvl) && stricmp(ni->nick, source) != 0) {
			ml = find_memolist(ni->nick);
			if (!ml) {
			    ml = scalloc(sizeof(MemoList), 1);
			    strscpy(ml->nick, ni->nick, NICKMAX);
			    alpha_insert_memolist(ml);
			}
			ml->n_memos++;
			ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
			m = &ml->memos[ml->n_memos-1];
			strscpy(m->sender, source, NICKMAX);
			if (ml->n_memos > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int i;
				for (i = 0; i < ml->n_memos; ++i)
				    ml->memos[i].number = i+1;
			    }
			} else
			    ml->memos[ml->n_memos-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(text);
		      m->flags = MF_UNREAD | MF_T_ADMIN | MF_L_INBOX;
			if ((ni->flags & NI_MEMO_RECEIVE) && finduser(ni->nick)) {
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_HAVE_NEW_MEMO_INBOX), source);
			    notice(s_MemoServ, ni->nick, rpl_numeric_lang(ni->nick, MS_PLEASE_TYPE_READ_IT), s_MemoServ, "INBOX", m->number);
		   }
	      }
	   }
      }
	notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT_ADMIN), lvl);
   }
}

/*************************************************************************/

/* ms_csend()
 * Send a memo to a channel. */

static void ms_csend(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    ChannelInfo *ci;
    CMemoList *cml;
    Memo *m;
    User *u;
    time_t now = time(NULL);

    crash_id = 74;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!text) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "CSEND <channel> <message>");
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP CSEND");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (!check_access(u, ci, CA_SEND_MEMO)) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
    } else if (conf.mssnddelay > 0 && ((u = finduser(source)) && u->lastmemosend + conf.mssnddelay > now)) {
	u->lastmemosend = now;
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_PLEASE_WAIT_SEC), conf.mssnddelay);
    } else {
	u->lastmemosend = now;
	cml = find_cmemolist(chan);
	if (!cml) {
	    cml = scalloc(sizeof(CMemoList), 1);
	    strscpy(cml->chan, chan, CHANMAX);
	    alpha_insert_cmemolist(cml);
	}
	cml->n_memos++;
	cml->memos = srealloc(cml->memos, sizeof(Memo) * cml->n_memos);
	m = &cml->memos[cml->n_memos-1];
	strscpy(m->sender, source, NICKMAX);
	if (cml->n_memos > 1) {
	    m->number = m[-1].number + 1;
	    if (m->number < 1) {
		int i;
		for (i = 0; i < cml->n_memos; ++i)
		    cml->memos[i].number = i+1;
	    }
	} else
	    cml->memos[cml->n_memos-1].number = 1;
	m->time = time(NULL);
	m->text = sstrdup(text);
      m->flags |= (MF_UNREAD | MF_T_CHANNEL | MF_L_INBOX);
      notice(s_MemoServ, source, rpl_numeric_lang(source, MS_MEMO_SENT), ci->name);
      /* Cycle the channel user list and notice each client. */
   }
}

/*************************************************************************/

/* ms_clist()
 * List the memos (if any) for the channel from the target folder, or all folders. */

static void ms_clist(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *param = strtok(NULL, " ");
    ChannelInfo *ci;
    CMemoList *cml;
    Memo *m;
    User *u;
    int i, inbox_cnt = 0, trash_cnt = 0, show_new = 0;
    char timebuf[64];
    struct tm tm;

    crash_id = 75;

    if (param && stricmp(param, "NEW") == 0) show_new = 1;

    if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (folder && chan) {
	if (stricmp(folder, "INBOX") == 0) {
		if (!(ci = cs_findchan(chan))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
			return;
		}
      	if (!check_access(u, ci, CA_READ_MEMO)) {
		   notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
		   return;
      	}
		if (!(cml = find_cmemolist(chan))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_CHAN_HAVE_NO_MEMOS), ci->name);
			return;
		}
		if (show_new) {
		    for (i = 0, m = cml->memos; i < cml->n_memos; i++, m++) {
			if ((m->flags & MF_UNREAD) && (m->flags & MF_L_INBOX))
			    break;
		    }
		    if (i == cml->n_memos) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_CHAN_HAVE_NO_MEMOS_INBOX), ci->name);
			return;
		    }
		}
		if (param)
			notice(s_MemoServ, source,
				"New memos in folder inbox for %s.  To read a memo in this channel's inbox, type \2/msg %s CREAD %s INBOX <number>.\2",
				ci->name, s_MemoServ, ci->name);
		else
			notice(s_MemoServ, source,
				"Memos in folder inbox for %s.  To read a memo in this channel's inbox folder, type \2/msg %s CREAD %s INBOX <number>.\2",
				ci->name, s_MemoServ, ci->name);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = cml->memos; i < cml->n_memos; ++i, ++m) {
		    if (m->flags & MF_L_INBOX) {
			    if (param && !(m->flags & MF_UNREAD))
				continue;
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
		             notice(s_MemoServ, source, "%3d %c %4d %-16s  %s", m->number, 
		                (m->flags & MF_UNREAD) ? 'U' : 'R', strlen(m->text), m->sender, timebuf);
			}
	 	} 
    	} else if (stricmp(folder, "TRASH") == 0) {
		if (!(ci = cs_findchan(chan))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
			return;
		}
      	if (!check_access(u, ci, CA_READ_MEMO)) {
		   notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
		   return;
      	}
		if (!(cml = find_cmemolist(chan))) {
			notice(s_MemoServ, source, rpl_numeric_lang(source, MS_CHAN_HAVE_NO_MEMOS), ci->name);
			return;
		}
		notice(s_MemoServ, source,
			"Memos in folder trash for %s.  To read a memo in this channel's trash folder, type \2/msg %s CREAD %s INBOX <number>.\2",
			ci->name, s_MemoServ, ci->name);
		notice(s_MemoServ, source, "Num S Size Sender            Date");
		for (i = 0, m = cml->trash; i < cml->n_trash; ++i, ++m) {
		    if (m->flags & MF_L_TRASH) {
			    tm = *localtime(&m->time);
			    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
			    timebuf[sizeof(timebuf)-1] = 0;
			    /* There are no new memos in this folder, as they are trash and we do not monitor it, simply
			     * list the memos, and be done.  Status - D relates to Deleted. */
		             notice(s_MemoServ, source, "%3d D %4d %-16s  %s", m->number, 
		                strlen(m->text), m->sender, timebuf);
			}
	 	} 
	} else {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "CLIST <channel> [\37INBOX\37|\37TRASH\37] [\37NEW\37]");
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP CLIST");
	}
    } else {
      if (!chan) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "CLIST <channel> [\37INBOX\37|\37TRASH\37] [\37NEW\37]");
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "HELP CLIST");
	} else if (!(ci = cs_findchan(chan))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
      } else if (!check_access(u, ci, CA_READ_MEMO)) {
	   	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
	} else if (!(cml = find_cmemolist(chan))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, MS_CHAN_HAVE_NO_MEMOS), ci->name);
	} else {
		notice(s_MemoServ, source, "Memo folders for channel %s.  To open a folder, type \2/msg %s CLIST %s <folder>.\2",
			ci->name, s_MemoServ, ci->name);
		for (i = 0, m = cml->memos; i < cml->n_memos; ++i, ++m) if (m->flags & MF_L_INBOX) inbox_cnt++;
		for (i = 0, m = cml->trash; i < cml->n_trash; ++i, ++m) if (m->flags & MF_L_TRASH) trash_cnt++;
		notice(s_MemoServ, source, "InBox (%d)     Trash (%d)", inbox_cnt, trash_cnt);
	}
   }
}

/*************************************************************************/

/* ms_cread()
 * Read a memo in the target folder. */

static void ms_cread(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    CMemoList *cml;
    ChannelInfo *ci;
    Memo *m;
    User *u;
    int num, i, show_last = 0;

    crash_id = 76;

    if (numstr && stricmp(numstr, "LAST") == 0) show_last = 1;
 
    if (!chan || !folder || (numstr && stricmp(numstr, "LAST") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: CREAD <channel> [INBOX|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP CREAD\2.", s_MemoServ);
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (stricmp(folder, "INBOX") == 0) {
	if (!(ci = cs_findchan(chan))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
		return;
	}
     	if (!check_access(u, ci, CA_READ_MEMO)) {
	   notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
	   return;
     	}
    	if (!(cml = find_cmemolist(chan))) {
		notice(s_MemoServ, source, "Channel \2%s\2 has no memos.", chan);
		return;
	}
	if (show_last) {
	    for (i = 0; i < cml->n_memos-1 && (cml->memos[i].flags & MF_L_INBOX); i++)
		;
	} else {
	    for (i = 0; i < cml->n_memos; i++) {
		if (cml->memos[i].number == num && (cml->memos[i].flags & MF_L_INBOX))
		    break;
	    }
	}
	if (i >= cml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d in channel \2%s\2 inbox folder does not exist.", num, ci->name);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &cml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in channel %s inbox folder from %s (%s).  To delete this memo in channel %s inbox folder, please type \2/msg %s CDEL %s INBOX %d\2.",
		m->number, ci->name, m->sender, timebuf, ci->name, s_MemoServ, ci->name, m->number);
	    if (m->flags & MF_T_SYS)
		    notice(s_MemoServ, source, "Memo type: System");
	    if (m->flags & MF_T_CHANNEL)
		    notice(s_MemoServ, source, "Memo type: Channel");
	    if (m->flags & MF_T_REPLY)
		    notice(s_MemoServ, source, "Memo type: Replied");
	    if (m->flags & MF_T_FORWARD)
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    notice(s_MemoServ, source, "%s", m->text);
	    m->flags &= ~MF_UNREAD;
	}
    } else if (stricmp(folder, "TRASH") == 0) {
	if (!(ci = cs_findchan(chan))) {
		notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
		return;
	}
     	if (!check_access(u, ci, CA_READ_MEMO)) {
	   notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
	   return;
     	}
    	if (!(cml = find_cmemolist(chan))) {
		notice(s_MemoServ, source, "Channel \2%s\2 has no memos.", chan);
		return;
	}
	if (show_last) {
	    for (i = 0; i < cml->n_trash-1 && (cml->trash[i].flags & MF_L_TRASH); i++)
		;
	} else {
	    for (i = 0; i < cml->n_trash; i++) {
		if (cml->trash[i].number == num && (cml->trash[i].flags & MF_L_TRASH))
		    break;
	    }
	}
	if (i >= cml->n_trash) {
	    notice(s_MemoServ, source, "Memo %d in channel \2%s\2 trash folder does not exist.", num, ci->name);
	} else {
	    char timebuf[64];
	    struct tm tm;
	    m = &cml->trash[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    notice(s_MemoServ, source, "Memo %d in channel %s trash folder from %s (%s).  To delete this memo in channel %s trash folder, please type \2/msg %s CDEL %s TRASH %d\2.",
		m->number, ci->name, m->sender, timebuf, ci->name, s_MemoServ, ci->name, m->number);
	    if (m->flags & MF_T_SYS)
		    notice(s_MemoServ, source, "Memo type: System");
	    if (m->flags & MF_T_CHANNEL)
		    notice(s_MemoServ, source, "Memo type: Channel");
	    if (m->flags & MF_T_REPLY)
		    notice(s_MemoServ, source, "Memo type: Replied");
	    if (m->flags & MF_T_FORWARD)
		    notice(s_MemoServ, source, "Memo type: Forwarded");
	    notice(s_MemoServ, source, "%s", m->text);
	    m->flags &= ~MF_UNREAD;
	}
    } else {
	notice(s_MemoServ, source, "Syntax: CREAD <channel> [INBOX|TRASH] [LAST] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP CREAD\2.", s_MemoServ);
    }
}

/*************************************************************************/

/* ms_cdel()
 * Delete a memo. */

static void ms_cdel(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *folder = strtok(NULL, " ");
    char *numstr = strtok(NULL, " ");
    CMemoList *cml;
    ChannelInfo *ci;
    Memo *m;
    User *u;
    int num, i, del_all = 0;

    crash_id = 77;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }

    if (numstr && stricmp(numstr, "ALL") == 0) del_all = 1;
 
    if (!folder || !chan || (numstr && stricmp(numstr, "ALL") != 0 && (num = atoi(numstr)) <= 0)) {
	notice(s_MemoServ, source, "Syntax: CDEL <channel> [INBOX|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP CDEL\2.", s_MemoServ);
    } else if (!(u = finduser(source))) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
    } else if (stricmp(folder, "INBOX") == 0) {
	    if (!(ci = cs_findchan(chan))) {
		 notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
		 return;
	    }
	    if (!check_access(u, ci, CA_DEL_MEMO)) {
		 notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
		 return;
	    }
	    if (!(cml = find_cmemolist(chan))) {
		notice(s_MemoServ, source, "Channel \2%s\2 has no memos.", ci->name);
		return;
	    }
          if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < cml->n_memos; ++i) {
			if (cml->memos[i].number == num)
			    break;
		    }
			/* Send this memo, to the trash folder. */
			cml->n_trash++;
			cml->trash = srealloc(cml->trash, sizeof(Memo) * cml->n_trash);
			m = &cml->trash[cml->n_trash-1];
			strscpy(m->sender, cml->memos[i].sender, NICKMAX);
			if (cml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < cml->n_trash; ++j)
				    cml->trash[j].number = j+1;
			    }
			} else
			    cml->trash[cml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(cml->memos[i].text);
		      m->flags |= cml->memos[i].flags;
			m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags |= MF_L_TRASH;

		    if (i < cml->n_memos) {
			free(cml->memos[i].text); 		/* Deallocate memo text memory */
			cml->n_memos--;		 		/* One less memo now */
			if (i < cml->n_memos)	 		/* Move remaining memos down a slot */
			    bcopy(cml->memos + i+1, cml->memos + i, sizeof(Memo) * (cml->n_memos - i));
			notice(s_MemoServ, source, "Memo %d in channel \2%s\2 inbox folder has been deleted.", num, ci->name);
		    } else {
			notice(s_MemoServ, source, "Memo %d in channel \2%s\2 inbox folder does not exist.", num, ci->name);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < cml->n_memos; ++i) {
			/* Send this memo, to the trash folder. */
			cml->n_trash++;
			cml->trash = srealloc(cml->trash, sizeof(Memo) * cml->n_trash);
			m = &cml->trash[cml->n_trash-1];
			strscpy(m->sender, cml->memos[i].sender, NICKMAX);
			if (cml->n_trash > 1) {
			    m->number = m[-1].number + 1;
			    if (m->number < 1) {
				int j;
				for (j = 0; j < cml->n_trash; ++j)
				    cml->trash[j].number = j+1;
			    }
			} else
			    cml->trash[cml->n_trash-1].number = 1;
			m->time = time(NULL);
			m->text = sstrdup(cml->memos[i].text);
		      m->flags |= cml->memos[i].flags;
			m->flags &= ~(MF_L_INBOX | MF_UNREAD);
			m->flags |= MF_L_TRASH;

			free(cml->memos[i].text);
		    }
		    cml->n_memos = 0;
		    notice(s_MemoServ, source, "All of the memos in channel \2%s\2 inbox folder have been deleted.", ci->name);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (cml->n_memos == 0 && cml->n_trash == 0)
		    del_cmemolist(cml);
    } else if (stricmp(folder, "TRASH") == 0) {
	    if (!(ci = cs_findchan(chan))) {
		 notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHAN_NOT_REGISTERED), chan);
		 return;
	    }
	    if (!check_access(u, ci, CA_DEL_MEMO)) {
		 notice(s_MemoServ, source, rpl_numeric_lang(source, CR_ACCESS_DENIED));
		 return;
	    }
	    if (!(cml = find_cmemolist(chan))) {
		notice(s_MemoServ, source, "Channel \2%s\2 has no memos.", ci->name);
		return;
	    }
		if (num > 0) {
		    /* Delete a specific memo. */	
		    for (i = 0; i < cml->n_trash; ++i) {
			if (cml->trash[i].number == num)
			    break;
		    }
		    if (i < cml->n_trash) {
			free(cml->trash[i].text); 		/* Deallocate memo text memory */
			cml->n_trash--;		 		/* One less memo now */
			if (i < cml->n_trash)	 		/* Move remaining memos down a slot */
			    bcopy(cml->trash + i+1, cml->trash + i, sizeof(Memo) * (cml->n_trash - i));
			notice(s_MemoServ, source, "Memo %d in channel \2%s\2 trash folder has been deleted.", num, ci->name);
		    } else {
			notice(s_MemoServ, source, "Memo %d in channel \2%s\2 trash folder does not exist.", num, ci->name);
		    }
		} 
		if (del_all) {
		    /* Delete all memos.  This requires freeing the memory holding
		     * the text of each memo and flagging that there are no memos
		     * left. */
		    for (i = 0; i < cml->n_trash; ++i)
			free(cml->trash[i].text);
		    cml->n_trash = 0;
		    notice(s_MemoServ, source, "All of memos in channel \2%s\2 trash folder have been deleted.", ci->name);
		}
		/* Did we delete the last memo?  If so, delete this MemoList. */
		if (cml->n_memos == 0 && cml->n_trash == 0)
		    del_cmemolist(cml);
    } else {
	notice(s_MemoServ, source, "Syntax: CDEL <channel> [INBOX|TRASH] [ALL] <\37number\37>");
	notice(s_MemoServ, source, "For more information please type \2/msg %s HELP CDEL\2.", s_MemoServ);
    }
}

/*************************************************************************/

#endif

/*************************************************************************/
