/* Routines for sending data to the network.
 * src/send.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef SEND_C
#define SEND_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

/* send_cmd()
 * Send a command to the server.  The two forms here are like
 * printf()/vprintf() and friends. */

void send_cmd(const char *source, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vsend_cmd(source, fmt, args);
    va_end(args);
}

/*************************************************************************/

/* vsend_cmd()
 * Send from send_cmd(). */

void vsend_cmd(const char *source, const char *fmt, va_list args)
{
    char buf[1024];	/* Big enough for IRC messages */

    vsnprintf(buf, sizeof(buf), fmt, args);
    if (source) {
	sockprintf(servsock, ":%s %s\r\n", source, buf);
	if (debug)
	    log("debug: Sent: :%s %s", source, buf);
    } else {
	sockprintf(servsock, "%s\r\n", buf);
	if (debug)
	    log("debug: Sent: %s", buf);
    }
}

/*************************************************************************/

/* wallops()
 * Send wallops to the network. */

void wallops(const char *source, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start(args, fmt);
    snprintf(buf, sizeof(buf), "WALLOPS :%s", fmt);
    vsend_cmd(source ? source : conf.lserver, buf, args);
}


/*************************************************************************/

/* globops()
 * Send globops to the network. */

void globops(const char *source, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start(args, fmt);
    snprintf(buf, sizeof(buf), "GLOBOPS :%s", fmt);
    vsend_cmd(source ? source : conf.lserver, buf, args);
}

/*************************************************************************/

/* notice()
 * Send a notice from the given source to the given nick. */

void notice(const char *source, const char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start(args, fmt);
    snprintf(buf, sizeof(buf), "NOTICE %s :%s", dest, fmt);
    vsend_cmd(source, buf, args);
}

/*************************************************************************/

/* notice_list()
 * Send a NULL-terminated array of text as NOTICEs. */

void notice_list(const char *source, const char *dest, const char **text)
{
    while (*text) {
	/* Have to kludge around an ircII bug here: if a notice includes
	 * no text, it is ignored, so we replace blank lines by lines
	 * with a single space.
	 */
	if (**text)
	    notice(source, dest, *text);
	else
	    notice(source, dest, " ");
	text++;
    }
}

/*************************************************************************/

/* privmsg()
 * Send a PRIVMSG from the given source to the given nick. */

void privmsg(const char *source, const char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start(args, fmt);
    snprintf(buf, sizeof(buf), "PRIVMSG %s :%s", dest, fmt);
    vsend_cmd(source, buf, args);
}

/*************************************************************************/

#endif

/*************************************************************************/